"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.parser.LipidBaseParserEventHandler import LipidBaseParserEventHandler
from pygoslin.domain.LipidAdduct import LipidAdduct
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.LipidCategory import LipidCategory
from pygoslin.domain.LipidClass import all_lipids
from pygoslin.domain.Adduct import Adduct
from pygoslin.domain.HeadGroup import HeadGroup
from pygoslin.domain.Element import element_positions

from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.Cycle import *

from pygoslin.domain.LipidCompleteStructure import LipidCompleteStructure
from pygoslin.domain.LipidFullStructure import LipidFullStructure
from pygoslin.domain.LipidStructureDefined import LipidStructureDefined
from pygoslin.domain.LipidSnPosition import LipidSnPosition
from pygoslin.domain.LipidMolecularSpecies import LipidMolecularSpecies
from pygoslin.domain.LipidSpecies import LipidSpecies
from pygoslin.domain.LipidSpeciesInfo import LipidSpeciesInfo

from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.LipidClass import *


class ShorthandParserEventHandler(LipidBaseParserEventHandler):
    
    def __init__(self):
        super().__init__()
        self.reset_lipid(None)
        
        self.registered_events["lipid_pre_event"] = self.reset_lipid
        self.registered_events["lipid_post_event"] = self.build_lipid
        
        ## set adduct events
        self.registered_events["adduct_info_pre_event"] = self.new_adduct
        self.registered_events["adduct_pre_event"] = self.add_adduct
        self.registered_events["charge_pre_event"] = self.add_charge
        self.registered_events["charge_sign_pre_event"] = self.add_charge_sign
        
        ## set species events
        self.registered_events["med_species_pre_event"] = self.set_species_level
        self.registered_events["gl_species_pre_event"] = self.set_species_level
        self.registered_events["gl_molecular_species_pre_event"] = self.set_molecular_level
        self.registered_events["pl_species_pre_event"] = self.set_species_level
        self.registered_events["pl_molecular_species_pre_event"] = self.set_molecular_level
        self.registered_events["sl_species_pre_event"] = self.set_species_level
        self.registered_events["la_species_pre_event"] = self.set_species_level
        
        self.registered_events["pl_single_pre_event"] = self.set_molecular_level
        self.registered_events["unsorted_fa_separator_pre_event"] = self.set_molecular_level
        self.registered_events["ether_num_pre_event"] = self.set_ether_num
        
        ## set head groups events
        self.registered_events["med_hg_single_pre_event"] = self.set_headgroup_name
        self.registered_events["med_hg_double_pre_event"] = self.set_headgroup_name
        self.registered_events["med_hg_triple_pre_event"] = self.set_headgroup_name
        self.registered_events["gl_hg_single_pre_event"] = self.set_headgroup_name
        self.registered_events["gl_hg_double_pre_event"] = self.set_headgroup_name
        self.registered_events["gl_hg_glycosyl_single_pre_event"] = self.set_headgroup_name
        self.registered_events["gl_hg_glycosyl_double_pre_event"] = self.set_headgroup_name
        self.registered_events["gl_hg_triple_pre_event"] = self.set_headgroup_name
        self.registered_events["pl_hg_single_pre_event"] = self.set_headgroup_name
        self.registered_events["pl_cpa_head_pre_event"] = self.set_headgroup_name
        self.registered_events["pl_hg_double_pre_event"] = self.set_headgroup_name
        self.registered_events["pl_hg_quadro_pre_event"] = self.set_headgroup_name
        self.registered_events["sl_hg_single_pre_event"] = self.set_headgroup_name
        self.registered_events["pl_hg_double_fa_hg_pre_event"] = self.set_headgroup_name
        self.registered_events["sl_hg_double_name_pre_event"] = self.set_headgroup_name
        self.registered_events["st_hg_pre_event"] = self.set_headgroup_name
        self.registered_events["st_hg_ester_pre_event"] = self.set_headgroup_name
        self.registered_events["hg_pip_pure_m_pre_event"] = self.set_headgroup_name
        self.registered_events["hg_pip_pure_d_pre_event"] = self.set_headgroup_name
        self.registered_events["hg_pip_pure_t_pre_event"] = self.set_headgroup_name
        self.registered_events["sl_hg_glyco_pre_event"] = self.set_headgroup_name
        self.registered_events["hg_PE_PS_pre_event"] = self.set_headgroup_name
        self.registered_events["la2_hg_pre_event"] = self.set_headgroup_name
        self.registered_events["la3_hg_pre_event"] = self.set_headgroup_name
        self.registered_events["la4_hg_pre_event"] = self.set_headgroup_name
        self.registered_events["acer_hg_post_event"] = self.set_acer
        self.registered_events["acer_species_post_event"] = self.set_acer_species
        self.registered_events["glyco_sphingo_lipid_pre_event"] = self.set_glyco_sphingo_lipid
        self.registered_events["carbohydrate_number_pre_event"] = self.set_carbohydrate_number

        ## set head group headgroup_decorators
        self.registered_events["carbohydrate_sn_pre_event"] = self.set_carbohydrate
        self.registered_events["carbohydrate_iso_pre_event"] = self.set_carbohydrate
        self.registered_events["carbohydrate_sn_position_pre_event"] = self.set_carbohydrate_sn_position
        self.registered_events["carbohydrate_isomeric_pre_event"] = self.set_carbohydrate_isomeric
        
        # fatty acyl events
        self.registered_events["lcb_pre_event"] = self.new_lcb
        self.registered_events["lcb_post_event"] = self.add_fatty_acyl_chain
        self.registered_events["fatty_acyl_chain_pre_event"] = self.new_fatty_acyl_chain
        self.registered_events["fatty_acyl_chain_post_event"] = self.add_fatty_acyl_chain
        self.registered_events["carbon_pre_event"] = self.set_carbon
        self.registered_events["db_count_pre_event"] = self.set_double_bond_count
        self.registered_events["db_position_number_pre_event"] = self.set_double_bond_position
        self.registered_events["db_single_position_pre_event"] = self.set_double_bond_information
        self.registered_events["db_single_position_post_event"] = self.add_double_bond_information
        self.registered_events["cistrans_pre_event"] = self.set_cistrans
        self.registered_events["ether_type_pre_event"] = self.set_ether_type
        self.registered_events["stereo_type_fa_pre_event"] = self.set_fatty_acyl_stereo
        
        ## set functional group events
        self.registered_events["func_group_data_pre_event"] = self.set_functional_group
        self.registered_events["func_group_data_post_event"] = self.add_functional_group
        self.registered_events["func_group_pos_number_pre_event"] = self.set_functional_group_position
        self.registered_events["func_group_name_pre_event"] = self.set_functional_group_name
        self.registered_events["func_group_count_pre_event"] = self.set_functional_group_count
        self.registered_events["stereo_type_fg_pre_event"] = self.set_functional_group_stereo
        self.registered_events["molecular_func_group_name_pre_event"] = self.set_sn_position_func_group
        self.registered_events["fa_db_only_post_event"] = self.add_dihydroxyl
        
        ## set cycle events
        self.registered_events["func_group_cycle_pre_event"] = self.set_cycle
        self.registered_events["func_group_cycle_post_event"] = self.add_cycle
        self.registered_events["cycle_start_pre_event"] = self.set_cycle_start
        self.registered_events["cycle_end_pre_event"] = self.set_cycle_end
        self.registered_events["cycle_number_pre_event"] = self.set_cycle_number
        self.registered_events["cycle_db_cnt_pre_event"] = self.set_cycle_db_count
        self.registered_events["cycle_db_positions_pre_event"] = self.set_cycle_db_positions
        self.registered_events["cycle_db_positions_post_event"] = self.check_cycle_db_positions
        self.registered_events["cycle_db_position_number_pre_event"] = self.set_cycle_db_position
        self.registered_events["cycle_db_position_cis_trans_pre_event"] = self.set_cycle_db_position_cistrans
        self.registered_events["cylce_element_pre_event"] = self.add_cycle_element
        
        ## set linkage events
        self.registered_events["fatty_acyl_linkage_pre_event"] = self.set_acyl_linkage
        self.registered_events["fatty_acyl_linkage_post_event"] = self.add_acyl_linkage
        self.registered_events["fatty_alkyl_linkage_pre_event"] = self.set_alkyl_linkage
        self.registered_events["fatty_alkyl_linkage_post_event"] = self.add_alkyl_linkage
        self.registered_events["fatty_linkage_number_pre_event"] = self.set_fatty_linkage_number
        self.registered_events["fatty_acyl_linkage_sign_pre_event"] = self.set_linkage_type
        self.registered_events["hydrocarbon_chain_pre_event"] = self.set_hydrocarbon_chain
        self.registered_events["hydrocarbon_chain_post_event"] = self.add_hydrocarbon_chain
        self.registered_events["hydrocarbon_number_pre_event"] = self.set_fatty_linkage_number
        
        ## set remaining events
        self.registered_events["ring_stereo_pre_event"] = self.set_ring_stereo
        self.registered_events["pl_hg_fa_pre_event"] = self.set_hg_acyl
        self.registered_events["pl_hg_fa_post_event"] = self.add_hg_acyl
        self.registered_events["pl_hg_alk_pre_event"] = self.set_hg_alkyl
        self.registered_events["pl_hg_alk_post_event"] = self.add_hg_alkyl
        self.registered_events["pl_hg_species_pre_event"] = self.add_pl_species_data
        self.registered_events["hg_pip_m_pre_event"] = self.suffix_decorator_molecular
        self.registered_events["hg_pip_d_pre_event"] = self.suffix_decorator_molecular
        self.registered_events["hg_pip_t_pre_event"] = self.suffix_decorator_molecular
        self.registered_events["hg_PE_PS_type_pre_event"] = self.suffix_decorator_species
        
        self.registered_events["sterol_definition_post_event"] = self.set_sterol_definition
        self.registered_events["sterol_ester_definition_post_event"] = self.set_sterol_definition
        self.registered_events["adduct_heavy_element_pre_event"] = self.set_heavy_element
        self.registered_events["adduct_heavy_number_pre_event"] = self.set_heavy_number
        self.registered_events["adduct_heavy_component_post_event"] = self.add_heavy_component
        self.debug = ""


    def reset_lipid(self, node):
        self.level = LipidLevel.COMPLETE_STRUCTURE
        self.head_group = ""
        self.adduct = None
        self.fa_list = []
        self.current_fa = []
        self.headgroup_decorators = []
        self.tmp = {}
        self.acer_species = False
        self.contains_stereo_information = False
        self.heavy_element = None
        self.heavy_element_number = 0
        
        
    
    def set_sterol_definition(self, node):
        self.head_group += " " + node.get_text()
    
    
        
    def add_cycle_element(self, node):
        element = node.get_text()
        if element not in element_positions:
            raise LipidParsingException("Element '%s' unknown" % element);
        self.tmp["fa%i" % len(self.current_fa)]["cycle_elements"].append(element_positions[element])
        
        
        
    def set_acer(self, node):
        self.head_group = "ACer"
        hgd = HeadgroupDecorator("decorator_acyl", suffix = True)
        hgd.functional_groups["decorator_acyl"] = [self.fa_list.pop()]
        self.headgroup_decorators.append(hgd)
        
        
        
    def set_acer_species(self, node):
        self.head_group = "ACer"
        self.set_lipid_level(LipidLevel.SPECIES)
        hgd = HeadgroupDecorator("decorator_acyl", suffix = True)
        hgd.functional_groups["decorator_acyl"] = [FattyAcid("FA", 2)]
        self.headgroup_decorators.append(hgd)
        self.acer_species = True
        
    
        
    def set_headgroup_name(self, node):
        if len(self.head_group) == 0: self.head_group = node.get_text()
        
        
        
    def set_carbohydrate_number(self, node):
        carbohydrate_num = int(node.get_text())
        if len(self.headgroup_decorators) and carbohydrate_num > 0:
            self.headgroup_decorators[-1].count += (carbohydrate_num - 1)
        
        
        
    def set_carbohydrate(self, node):
        carbohydrate = node.get_text()
        try:
            functional_group = get_functional_group(carbohydrate)
        except Exception:
            raise LipidParsingException("Carbohydrate '%s' unknown" % carbohydrate)
        
        functional_group.elements[Element.O] -= 1
        if "func_group_head" in self.tmp and self.tmp["func_group_head"]:
            self.headgroup_decorators.append(functional_group)
        else:
            if carbohydrate not in self.current_fa[-1].functional_groups:
                self.current_fa[-1].functional_groups[carbohydrate] = []
            self.current_fa[-1].functional_groups[carbohydrate].append(functional_group)
        
        
        
    def set_carbohydrate_sn_position(self, node):
        self.set_lipid_level(LipidLevel.SN_POSITION)
        self.tmp["func_group_head"] = True
        
        
        
    def set_carbohydrate_isomeric(self, node):
        self.tmp["func_group_head"] = True
        
        
        
    def suffix_decorator_molecular(self, node):
        self.headgroup_decorators.append(HeadgroupDecorator(node.get_text(), suffix = True, level = LipidLevel.MOLECULAR_SPECIES))
        
        
        
    def suffix_decorator_species(self, node):
        self.headgroup_decorators.append(HeadgroupDecorator(node.get_text(), suffix = True, level = LipidLevel.SPECIES))
        
        
    
    def set_pl_hg_triple(self, node):
        self.set_molecular_level(node)
        self.set_headgroup_name(node)
        
        
        
    def set_ring_stereo(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["fg_ring_stereo"] = node.get_text()
        
        
        
    def set_lcb(self, node):
        self.fa_list[-1].set_type(LipidFaBondType.LCB_REGULAR)
        
        
        
    def add_pl_species_data(self, node):
        self.set_lipid_level(LipidLevel.SPECIES)
        hgd = HeadgroupDecorator("")
        hgd.elements[Element.O] += 1
        hgd.elements[Element.H] -= 1
        self.headgroup_decorators.append(hgd)
    
        
    
    def new_fatty_acyl_chain(self, node):
        self.current_fa.append(FattyAcid("FA"))
        self.tmp["fa%i" % len(self.current_fa)] = {}
    
    
    
    def new_lcb(self, node):
        self.new_fatty_acyl_chain(node)
        self.current_fa[-1].set_type(LipidFaBondType.LCB_REGULAR)
        
        
        
    def add_fatty_acyl_chain(self, node):
        fg_i = "fa%i" % (len(self.current_fa) - 2)
        special_type = ""
        if len(self.current_fa) >= 2 and fg_i in self.tmp and "fg_name" in self.tmp[fg_i]:
            if self.tmp[fg_i]["fg_name"] in {"acyl", "alkyl", "decorator_acyl", "decorator_alkyl", "cc"}:
                special_type = self.tmp[fg_i]["fg_name"]
                
        
        fa_i = "fa%i" % len(self.current_fa)
        if type(self.current_fa[-1].double_bonds) != int:
            if len(self.current_fa[-1].double_bonds) != self.tmp[fa_i]["db_count"]:
                raise LipidException("Double bond count does not match with number of double bond positions")
            
        else:
            if self.current_fa[-1].double_bonds > 0:
                self.set_lipid_level(LipidLevel.SN_POSITION)
                
        del self.tmp[fa_i]
        
        if len(special_type) > 0:
            fg_acyl_alkyl = self.current_fa.pop()
            fg_acyl_alkyl.name = special_type
            if special_type not in self.current_fa[-1].functional_groups: self.current_fa[-1].functional_groups[special_type] = []
            self.current_fa[-1].functional_groups[special_type].append(fg_acyl_alkyl)
            
        else:
            self.fa_list.append(self.current_fa.pop())
        
        
        
    def set_double_bond_position(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        self.tmp[fa_i]["db_position"] = int(node.get_text())
        
        
        
    def set_double_bond_information(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        self.tmp[fa_i]["db_position"] = 0
        self.tmp[fa_i]["db_cistrans"] = ""
        
        
        
    def add_double_bond_information(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        pos = self.tmp[fa_i]["db_position"]
        cistrans = self.tmp[fa_i]["db_cistrans"]
        
        if cistrans == "": self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        
        del self.tmp[fa_i]["db_position"]
        del self.tmp[fa_i]["db_cistrans"]
        if type(self.current_fa[-1].double_bonds) == int: self.current_fa[-1].double_bonds = {}
        self.current_fa[-1].double_bonds[pos] = cistrans
        
        
        
    def set_cistrans(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["db_cistrans"] = node.get_text()
        
        
    
    def set_fatty_acyl_stereo(self, node):
        self.current_fa[-1].stereochemistry = node.get_text()
        self.contains_stereo_information = True
        
        

    def add_dihydroxyl(self, node):
        if self.current_fa[-1].lipid_FA_bond_type not in FattyAcid.LCB_STATES: return
    
        num_h = 1
        functional_group = get_functional_group("OH")
        if self.head_group in self.SP_EXCEPTION_CLASSES and len(self.headgroup_decorators) == 0: num_h += 1
        
        functional_group.count = num_h
        if "OH" not in self.current_fa[-1].functional_groups: self.current_fa[-1].functional_groups["OH"] = []
        self.current_fa[-1].functional_groups["OH"].append(functional_group)
        
        
        
    def set_functional_group(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        self.tmp[fa_i]["fg_pos"] = -1
        self.tmp[fa_i]["fg_name"] = "O"
        self.tmp[fa_i]["fg_cnt"] = 1
        self.tmp[fa_i]["fg_stereo"] = ""
        self.tmp[fa_i]["fg_ring_stereo"] = ""
        
        
        
    def set_cycle(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["fg_name"] = "cy"
        self.current_fa.append(Cycle(0))
        fa_i = "fa%i" % len(self.current_fa)
        self.tmp[fa_i] = {"cycle_elements": []}
        
        
        
    def add_cycle(self, node):
        cycle_elements = self.tmp["fa%i" % len(self.current_fa)]["cycle_elements"]
        del self.tmp["fa%i" % len(self.current_fa)]
        cycle = self.current_fa.pop()
        cycle.bridge_chain = cycle_elements
        
        
        if type(cycle.start) == int and type(cycle.end) == int:
            if cycle.end - cycle.start + 1 + len(cycle.bridge_chain) < cycle.cycle:
                raise ConstraintViolationException("Cycle length '%i' does not match with cycle description" % cycle.cycle)
        
        if "cy" not in self.current_fa[-1].functional_groups: self.current_fa[-1].functional_groups["cy"] = []
        self.current_fa[-1].functional_groups["cy"].append(cycle)
        
        
        
    def set_fatty_linkage_number(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["linkage_pos"] = int(node.get_text())
        
        
        
    def set_hg_acyl(self, node):
        self.tmp["fa%i" % len(self.current_fa)] = {"fg_name": "decorator_acyl"}
        self.current_fa.append(HeadgroupDecorator("decorator_acyl", suffix = True))
        self.tmp["fa%i" % len(self.current_fa)] = {}
        
        
    
    def add_hg_acyl(self, node):
        del self.tmp["fa%i" % len(self.current_fa)]
        self.headgroup_decorators.append(self.current_fa.pop())
        del self.tmp["fa%i" % len(self.current_fa)]
        
        
        
    def set_glyco_sphingo_lipid(self, node):
        pass
        
        
        
    def set_hg_alkyl(self, node):
        self.tmp["fa%i" % len(self.current_fa)] = {"fg_name": "decorator_alkyl"}
        self.current_fa.append(HeadgroupDecorator("decorator_alkyl", suffix = True))
        self.tmp["fa%i" % len(self.current_fa)] = {}
        
    
    
    def add_hg_alkyl(self, node):
        del self.tmp["fa%i" % len(self.current_fa)]
        self.headgroup_decorators.append(self.current_fa.pop())
        del self.tmp["fa%i" % len(self.current_fa)]
        
        
        
    def set_linkage_type(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["linkage_type"] = node.get_text() == "N"
        
        
        
    def set_hydrocarbon_chain(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        self.tmp[fa_i]["fg_name"] = "cc"
        self.current_fa.append(CarbonChain(None))
        self.tmp["fa%i" % len(self.current_fa)] = {"linkage_pos": -1}
        
        
        
    def add_hydrocarbon_chain(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        linkage_pos = self.tmp[fa_i]["linkage_pos"]
        
        del self.tmp[fa_i]
        cc = self.current_fa.pop()
        
        cc.position = linkage_pos
        if linkage_pos == -1: self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        
        if "cc" not in self.current_fa[-1].functional_groups: self.current_fa[-1].functional_groups["cc"] = []
        self.current_fa[-1].functional_groups["cc"].append(cc)
        
        
        
    def set_acyl_linkage(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        self.tmp[fa_i]["fg_name"] = "acyl"
        self.current_fa.append(AcylAlkylGroup(None))
        self.tmp["fa%i" % len(self.current_fa)] = {"linkage_pos": -1}
        

        
    def add_acyl_linkage(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        linkage_type = self.tmp[fa_i]["linkage_type"]
        linkage_pos = self.tmp[fa_i]["linkage_pos"]
        
        del self.tmp[fa_i]
        acyl = self.current_fa.pop()
        
        acyl.position = linkage_pos
        acyl.set_N_bond_type(linkage_type)
        if linkage_pos == -1: self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        
        if "acyl" not in self.current_fa[-1].functional_groups: self.current_fa[-1].functional_groups["acyl"] = []
        self.current_fa[-1].functional_groups["acyl"].append(acyl)
        
        
        
    def set_alkyl_linkage(self, node):
        fa_i = "fa%i" % len(self.current_fa)
        self.tmp[fa_i]["fg_name"] = "alkyl"
        self.current_fa.append(AcylAlkylGroup(None, alkyl = True))
        self.tmp["fa%i" % len(self.current_fa)] = {"linkage_pos": -1}
        
        
        
    def add_alkyl_linkage(self, node):
        linkage_pos = self.tmp["fa%i" % len(self.current_fa)]["linkage_pos"]
        
        del self.tmp["fa%i" % len(self.current_fa)]
        alkyl = self.current_fa.pop()
        
        alkyl.position = linkage_pos
        if linkage_pos == -1: self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        
        if "alkyl" not in self.current_fa[-1].functional_groups: self.current_fa[-1].functional_groups["alkyl"] = []
        self.current_fa[-1].functional_groups["alkyl"].append(alkyl)
        
        
        
    def set_cycle_start(self, node):
        self.current_fa[-1].start = int(node.get_text())
        self.current_fa[-1].position = int(node.get_text())
        
        
        
    def set_cycle_end(self, node):
        self.current_fa[-1].end = int(node.get_text())
        
        
        
    def set_cycle_number(self, node):
        self.current_fa[-1].cycle = int(node.get_text())
        
        
        
    def set_cycle_db_count(self, node):
        self.current_fa[-1].double_bonds = int(node.get_text())
    
    
    
    def set_cycle_db_positions(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["cycle_db"] = self.current_fa[-1].double_bonds
        self.current_fa[-1].double_bonds = {}
    
    
    
    def check_cycle_db_positions(self, node):
        if len(self.current_fa[-1].double_bonds) != self.tmp["fa%i" % len(self.current_fa)]["cycle_db"]:
            raise LipidException("Double bond number in cycle does not correspond to number of double bond positions.")
    
    
    
    def set_cycle_db_position(self, node):
        pos = int(node.get_text())
        self.current_fa[-1].double_bonds[pos] = ""
        self.tmp["fa%i" % len(self.current_fa)]["last_db_pos"] = pos
        
        
        
    def set_cycle_db_position_cistrans(self, node):
        pos = self.tmp["fa%i" % len(self.current_fa)]["last_db_pos"]
        self.current_fa[-1].double_bonds[pos] = node.get_text()
        
    
    def set_functional_group_position(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["fg_pos"] = int(node.get_text())
        
    
    def set_functional_group_name(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["fg_name"] = node.get_text()
        
    
    
    def set_functional_group_count(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["fg_cnt"] = int(node.get_text())
        
    
    
    def set_functional_group_stereo(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["fg_stereo"] = node.get_text()
        self.contains_stereo_information = True
        
        
        
    def set_sn_position_func_group(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["fg_name"] = node.get_text()
        self.set_lipid_level(LipidLevel.SN_POSITION)
        
        
        
    def add_functional_group(self, node):
        
        fa_i = "fa%i" % len(self.current_fa)
        fg_pos = self.tmp[fa_i]["fg_pos"]
        fg_name = self.tmp[fa_i]["fg_name"]
        fg_cnt = self.tmp[fa_i]["fg_cnt"]
        fg_stereo = self.tmp[fa_i]["fg_stereo"]
        fg_ring_stereo = self.tmp[fa_i]["fg_ring_stereo"]
        
        if fg_name in {"cy", "cc", "acyl", "alkyl", "decorator_acyl", "decorator_alkyl"}: return
        
        if fg_pos == -1:
            self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
            
        if fg_cnt <= 0:
            return
        
        try:
            functional_group = get_functional_group(fg_name)
        except Exception:
            raise LipidParsingException(" '%s' unknown" % fg_name)
        
        functional_group.position = fg_pos
        functional_group.count = fg_cnt
        functional_group.stereochemistry = fg_stereo
        functional_group.ring_stereo = fg_ring_stereo
        
        del self.tmp[fa_i]["fg_pos"]
        del self.tmp[fa_i]["fg_name"]
        del self.tmp[fa_i]["fg_cnt"]
        del self.tmp[fa_i]["fg_stereo"]
        
        if fg_name not in self.current_fa[-1].functional_groups: self.current_fa[-1].functional_groups[fg_name] = []
        self.current_fa[-1].functional_groups[fg_name].append(functional_group)
        
        
        
    def set_ether_type(self, node):
        ether_type = node.get_text()
        if ether_type == "O-": self.current_fa[-1].lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMANYL
        elif ether_type == "P-": self.current_fa[-1].lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMENYL
        
        
    def set_ether_num(self, node):
        num_ethers, ether = 0, node.get_text()
        if ether == "d": num_ethers = 2
        elif ether == "t": num_ethers = 3
        elif ether == "e": num_ethers = 4
        self.tmp["num_ethers"] = num_ethers
        
        
    def set_species_level(self, node):
        self.set_lipid_level(LipidLevel.SPECIES)
        
        
        
    def set_molecular_level(self, node):
        self.set_lipid_level(LipidLevel.MOLECULAR_SPECIES)
        
        
    
    def set_carbon(self, node):
        self.current_fa[-1].num_carbon = int(node.get_text())
      
      
      
    def set_double_bond_count(self, node):
        self.tmp["fa%i" % len(self.current_fa)]["db_count"] = self.current_fa[-1].double_bonds = int(node.get_text())
        
        
        
    def build_lipid(self, node):
        if self.acer_species: self.fa_list[0].num_carbon -= 2
        headgroup = self.prepare_headgroup_and_checks(False)
        
        lipid = LipidAdduct()
        lipid.adduct = self.adduct
        lipid.lipid = self.assemble_lipid(headgroup)
            
        if "num_ethers" in self.tmp: lipid.lipid.info.num_ethers = self.tmp["num_ethers"]
        
        self.content = lipid
        
        
        
    def new_adduct(self, node):
        if self.adduct == None: self.adduct = Adduct("", "")
        
        
        
    def add_adduct(self, node):
        self.adduct.adduct_string = node.get_text()
        
        
        
    def add_charge(self, node):
        self.adduct.charge = int(node.get_text())
        
        
        
    def add_charge_sign(self, node):
        sign = node.get_text()
        if sign == "+": self.adduct.set_charge_sign(1)
        if sign == "-": self.adduct.set_charge_sign(-1)
        if self.adduct.charge == 0: self.adduct.charge = 1
        
        
        
    def set_heavy_element(self, node):
        self.heavy_element = heavy_element_table[node.get_text()]
        self.heavy_element_number = 1
        
        
        
    def set_heavy_number(self, node):
        self.heavy_element_number = int(node.get_text())
        
        
        
    def add_heavy_component(self, node):
        self.adduct.heavy_elements[self.heavy_element] = self.heavy_element_number
        



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""



import pygoslin
from os import path
from pygoslin.parser.GoslinParserEventHandler import GoslinParserEventHandler
from pygoslin.parser.LipidMapsParserEventHandler import LipidMapsParserEventHandler
from pygoslin.parser.SwissLipidsParserEventHandler import SwissLipidsParserEventHandler
from pygoslin.parser.HmdbParserEventHandler import HmdbParserEventHandler
from pygoslin.parser.ShorthandParserEventHandler import ShorthandParserEventHandler
from pygoslin.parser.FattyAcidParserEventHandler import FattyAcidParserEventHandler
#from pygoslin.parser.SystematicParserEventHandler import SystematicParserEventHandler

from pygoslin.parser.ParserCommon import Parser
from pygoslin.domain.LipidExceptions import LipidException

        
    
    
class GoslinParser(Parser):
    def __init__(self):
        self.event_handler = GoslinParserEventHandler()
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "Goslin.g4")
        super().__init__(self.event_handler, file_name, Parser.DEFAULT_QUOTE)
        
        

class ShorthandParser(Parser):
    def __init__(self):
        self.event_handler = ShorthandParserEventHandler()
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "Shorthand2020.g4")
        super().__init__(self.event_handler, file_name, Parser.DEFAULT_QUOTE)
        
        
"""
class SystematicParser(Parser):
    def __init__(self):
        self.event_handler = SystematicParserEventHandler()
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "Systematic.g4")
        super().__init__(self.event_handler, file_name, Parser.DEFAULT_QUOTE)
        
    def parse(self, lipid_name):
        return super().parse(lipid_name.lower())
"""
        
        
class FattyAcidParser(Parser):
    def __init__(self):
        self.event_handler = FattyAcidParserEventHandler()
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "FattyAcids.g4")
        super().__init__(self.event_handler, file_name, Parser.DEFAULT_QUOTE)
        
    def parse(self, lipid_name, raise_error = True):
        return super().parse(lipid_name.lower(), raise_error = raise_error)
        
        
class LipidMapsParser(Parser):
    def __init__(self):
        self.event_handler = LipidMapsParserEventHandler()
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "LipidMaps.g4")
        super().__init__(self.event_handler, file_name, Parser.DEFAULT_QUOTE)
        
        
class SwissLipidsParser(Parser):
    def __init__(self):
        self.event_handler = SwissLipidsParserEventHandler()
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "SwissLipids.g4")
        super().__init__(self.event_handler, file_name, Parser.DEFAULT_QUOTE)
        
        
        
        
class HmdbParser(Parser):
    def __init__(self):
        self.event_handler = HmdbParserEventHandler()
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "HMDB.g4")
        super().__init__(self.event_handler, file_name, Parser.DEFAULT_QUOTE)
        
        
        
class LipidParser:
    def __init__(self):
        self.parser_list = [ShorthandParser(), GoslinParser(), FattyAcidParser(), LipidMapsParser(), SwissLipidsParser(), HmdbParser()]
        self.parser_event_handler = None
        """
        self.trivial_names = {}
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "trivial-names.csv")
        
        with open(file_name, "rt", encoding= "utf-8") as trivial_infile:
            trivial_infile.readline()
            for line in trivial_infile:
                line = line.strip()
                if len(line) < 1: continue
                row = Parser.split_string(line.strip(), ",", '"')
                if len(row) < 2: continue
            
                value = row[0].strip('"')
                for key in row[1:]:
                    key = key.strip('"').lower()
                    if len(key) < 1: continue
                
                    if key in self.trivial_names:
                        print("Error: key %s in trivial names list occurs twice." % key)
                        exit()
                    self.trivial_names[key] = value
        """
                
        
    def parse(self, lipid_name):
        # check if lipid name has a trivial name and translate it via predefined table
        #lipid_name_lower = lipid_name.lower()
        #if lipid_name_lower in self.trivial_names:
        #    lipid_name = self.trivial_names[lipid_name_lower]
        
        self.parser_event_handler = None
        # go through all parsers
        for parser in self.parser_list:
            lipid = parser.parse(lipid_name, raise_error = False)
            if lipid != None:
                self.parser_event_handler = parser.parser_event_handler
                return lipid
            
        raise LipidException("Lipid not found")




"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.parser.BaseParserEventHandler import BaseParserEventHandler
from pygoslin.domain.LipidAdduct import LipidAdduct
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.LipidClass import *
from pygoslin.domain.Adduct import Adduct

from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.HeadGroup import HeadGroup, glyco_table
from pygoslin.domain.FunctionalGroup import *

from pygoslin.domain.LipidCompleteStructure import LipidCompleteStructure
from pygoslin.domain.LipidFullStructure import LipidFullStructure
from pygoslin.domain.LipidStructureDefined import LipidStructureDefined
from pygoslin.domain.LipidSnPosition import LipidSnPosition
from pygoslin.domain.LipidMolecularSpecies import LipidMolecularSpecies
from pygoslin.domain.LipidSpecies import LipidSpecies

from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.Cycle import *


class LipidBaseParserEventHandler(BaseParserEventHandler):
    SP_EXCEPTION_CLASSES = set(all_lipids[get_class("Cer")]["synonyms"]) | set(all_lipids[get_class("SPB")]["synonyms"]) | set(["Cer"]) | set(["SPB"])
    
    def __init__(self):
        super().__init__()
        self.level = LipidLevel.FULL_STRUCTURE
        self.head_group = ""
        self.lcb = None
        self.fa_list = []
        self.current_fa = None
        self.adduct = None
        self.headgroup_decorators = []
        self.use_head_group = False
        
        
        
    def check_full_structure(obj):
        full = True
        if type(obj) == FattyAcid and obj.num_carbon == 0: return True
        if type(obj) == FattyAcid and type(obj.double_bonds) == int and obj.double_bonds > 0: return False
        if type(obj) == FattyAcid and type(obj.double_bonds) == dict:
            full &= len(obj.double_bonds) == sum(1 if (v in {"E", "Z"} or (v == '' and k == obj.num_carbon - 1)) else 0 for k, v in obj.double_bonds.items())
        for fg_name in obj.functional_groups:
            for fg in obj.functional_groups[fg_name]:
                if fg.name == "X": continue
                if fg.position < 0: return False
                if obj.functional_groups != None: full &= LipidBaseParserEventHandler.check_full_structure(fg)
        return full
    
    
    def set_lipid_level(self, level):
        self.level = self.level if self.level.value < level.value else level
        
        
        
    def sp_regular_lcb(self):
        return get_category(self.head_group) == LipidCategory.SP and self.current_fa.lipid_FA_bond_type in FattyAcid.LCB_STATES and not (self.head_group in self.SP_EXCEPTION_CLASSES and len(self.headgroup_decorators) == 0)
    
    
    
    def prepare_headgroup_and_checks(self, allow_class_shift = True):
        
        headgroup = HeadGroup(self.head_group, self.headgroup_decorators, self.use_head_group)
        if self.use_head_group: return headgroup
    
        self.head_group = all_lipids[headgroup.lipid_class]["name"]
        max_num_fa = all_lipids[headgroup.lipid_class]["max_fa"]
        if max_num_fa != len(self.fa_list): self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
    
        true_fa = sum(1 for fa in self.fa_list if fa.num_carbon > 0 or (fa.double_bonds > 0 if type(fa.double_bonds) == int else len(fa.double_bonds)) > 0)
        
        poss_fa = all_lipids[headgroup.lipid_class]["poss_fa"]
        max_fa = all_lipids[headgroup.lipid_class]["max_fa"]
        
        # make lyso
        can_be_lyso = "Lyso" in all_lipids[get_class("L" + self.head_group)]["specials"] if get_class("L" + self.head_group) < len(all_lipids) else False
        
        if allow_class_shift:
            if (true_fa + 1 == poss_fa or true_fa + 2 == poss_fa) and self.level != LipidLevel.SPECIES and headgroup.lipid_category == LipidCategory.GP and can_be_lyso:
                if true_fa + 1 == poss_fa: self.head_group = "L" + self.head_group
                else: self.head_group = "DL" + self.head_group
                headgroup = HeadGroup(self.head_group, self.headgroup_decorators, self.use_head_group)
                poss_fa = all_lipids[headgroup.lipid_class]["poss_fa"] if headgroup.lipid_class < len(all_lipids) else 0
                 
            elif (true_fa + 1 == poss_fa or true_fa + 2 == poss_fa) and self.level != LipidLevel.SPECIES and headgroup.lipid_category == LipidCategory.GL and self.head_group == "TG":
                if true_fa + 1 == poss_fa: self.head_group = "DG"
                else: self.head_group = "MG"
                headgroup = HeadGroup(self.head_group, self.headgroup_decorators, self.use_head_group)
                poss_fa = all_lipids[headgroup.lipid_class]["poss_fa"] if headgroup.lipid_class < len(all_lipids) else 0


        # check if all functional groups have a position to be full structure
        if self.level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE}:
            for fa in self.fa_list:
                if not LipidBaseParserEventHandler.check_full_structure(fa):
                    self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
                    break
                
                
        if self.level == LipidLevel.MOLECULAR_SPECIES and headgroup.lipid_category == LipidCategory.GP and len(self.fa_list) == 2 and sum(fa.lipid_FA_bond_type == LipidFaBondType.ESTER for fa in self.fa_list) < 2:
            self.level = LipidLevel.SN_POSITION
            if self.fa_list[0].lipid_FA_bond_type == LipidFaBondType.ESTER:
                self.fa_list[0], self.fa_list[1] = self.fa_list[1], self.fa_list[0]
                    

        if self.level == LipidLevel.SPECIES:
            if true_fa == 0 and poss_fa != 0:
                raise ConstraintViolationException("No fatty acyl information lipid class '%s' provided." % headgroup.headgroup)
            
        elif true_fa > poss_fa and self.level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE, LipidLevel.STRUCTURE_DEFINED, LipidLevel.SN_POSITION, LipidLevel.MOLECULAR_SPECIES}:
            raise ConstraintViolationException("Number of specified fatty acyl chains (%i) not allowed for lipid class '%s' (having %i fatty aycl chains)." % (true_fa, headgroup.headgroup, poss_fa))
        
        elif "Lyso" in all_lipids[get_class(self.head_group)]["specials"] and true_fa > poss_fa:
            raise ConstraintViolationException("Number of specified fatty acyl chains (%i) not allowed for lipid class '%s' (having %i fatty aycl chains)." % (true_fa, headgroup.headgroup, poss_fa))
            
        
        if "HC" in all_lipids[headgroup.lipid_class]["specials"] and len(self.fa_list) > 0:
            self.fa_list[0].lipid_FA_bond_type = LipidFaBondType.ETHER
            
        if "Amide" in all_lipids[headgroup.lipid_class]["specials"] and len(self.fa_list) > 0:
            for fatty in self.fa_list: fatty.lipid_FA_bond_type = LipidFaBondType.AMIDE
        
        # make LBC exception
        if len(self.fa_list) > 0 and headgroup.sp_exception:
            self.fa_list[0].set_type(LipidFaBondType.LCB_EXCEPTION)
            
        return headgroup
    
    
    
    def assemble_lipid(self, headgroup):
        for fa in self.fa_list:
            if fa.stereo_information_missing():
                self.set_lipid_level(LipidLevel.FULL_STRUCTURE)
                break
        
        lipid_level_class = None
        if self.level == LipidLevel.COMPLETE_STRUCTURE: lipid_level_class = LipidCompleteStructure
        elif self.level == LipidLevel.FULL_STRUCTURE: lipid_level_class = LipidFullStructure
        elif self.level == LipidLevel.STRUCTURE_DEFINED: lipid_level_class = LipidStructureDefined
        elif self.level == LipidLevel.SN_POSITION: lipid_level_class = LipidSnPosition
        elif self.level == LipidLevel.MOLECULAR_SPECIES: lipid_level_class = LipidMolecularSpecies
        elif self.level == LipidLevel.SPECIES: lipid_level_class = LipidSpecies
        
        return lipid_level_class(headgroup, self.fa_list)
    
    
    
    
    
    def resolve_fa_synonym(self, mediator_name):
        
        if mediator_name == "Palmitic acid": # FA 16:0
            return FattyAcid("FA", 16)
            
        elif mediator_name in {"Linoleic acid", "LA"}:
            return FattyAcid("FA", 18, {9: "Z", 12: "Z"})
            
        elif mediator_name in {"Arachidonic acid", "AA", "ARA"}:
            return FattyAcid("FA", 20, {5: "Z", 8: "Z", 11: "Z", 14: "Z"})
            
        elif mediator_name == "ALA":
            return FattyAcid("FA", 18, {9: "Z", 12: "Z", 15: "Z"})
            
        elif  mediator_name == "EPA":
            return FattyAcid("FA", 20, {5: "Z", 8: "Z", 11: "Z", 14: "Z", 17: "Z"})
            
        elif  mediator_name == "DHA":
            return FattyAcid("FA", 22, {4: "Z", 7: "Z", 10: "Z", 13: "Z", 16: "Z", 19: "Z"})
            
        elif mediator_name == "LTB4":
            f1, f2 = get_functional_group("OH"), get_functional_group("OH")
            f1.position = 5
            f2.position = 12
            fg = {"OH": [f1, f2]}
            return FattyAcid("FA", 20, {6: "Z", 8: "E", 10: "E", 14: "Z"}, functional_groups = fg)
            
        elif mediator_name in {"Resolvin D3", "RvD3"}:
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH")
            f1.position = 4
            f2.position = 11
            f3.position = 17
            fg = {"OH": [f1, f2, f3]}
            return FattyAcid("FA", 22, {5: "E", 7: "E", 9: "E", 13: "Z", 15: "E", 19: "Z"}, functional_groups = fg)
            
        elif mediator_name in {"Maresin 1", "Mar1"}:
            f1, f2 = get_functional_group("OH"), get_functional_group("OH")
            f1.position = 7
            f2.position = 14
            fg = {"OH": [f1, f2]}
            return FattyAcid("FA", 22, {4: "Z", 8: "E", 10: "E", 12: "Z", 16: "Z", 19: "Z"}, functional_groups = fg)
            
        elif mediator_name == "Resolvin D1":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH")
            f1.position = 7
            f2.position = 8
            f3.position = 17
            fg = {"OH": [f1, f2, f3]}
            return FattyAcid("FA", 22, {4: "Z", 9: "E", 11: "E", 13: "Z", 15: "E", 19: "Z"}, functional_groups = fg)
            
        elif mediator_name == "Resolvin D2":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH")
            f1.position = 4
            f2.position = 16
            f3.position = 17
            fg = {"OH": [f1, f2, f3]}
            return FattyAcid("FA", 22, {4: "Z", 8: "E", 10: "Z", 12: "E", 14: "E", 19: "Z"}, functional_groups = fg)
            
        elif mediator_name == "Resolvin D5":
            f1, f2 = get_functional_group("OH"), get_functional_group("OH")
            f1.position = 7
            f2.position = 17
            fg = {"OH": [f1, f2]}
            return FattyAcid("FA", 22, {4: "Z", 8: "E", 10: "Z", 13: "Z", 15: "E", 19: "Z"}, functional_groups = fg)
            
        elif mediator_name == "Resolvin E1":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH")
            f1.position = 5
            f2.position = 13
            f3.position = 18
            fg = {"OH": [f1, f2, f3]}
            return FattyAcid("FA", 20, {6: "Z", 8: "E", 10: "E", 14: "Z", 16: "E"}, functional_groups = fg)
            
        elif mediator_name == "Resolvin E2":
            f1, f2 = get_functional_group("OH"), get_functional_group("OH")
            f1.position = 5
            f2.position = 18
            fg = {"OH": [f1, f2]}
            return FattyAcid("FA", 20, {6: "E", 8: "Z", 11: "Z", 14: "Z", 16: "E"}, functional_groups = fg)
            
        elif mediator_name == "TXB1":
            f1, f2, f3, f4 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH"), get_functional_group("oxy")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            f4.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f2, f3], "oxy": [f4]})]}
            return FattyAcid("FA", 20, {13: "E"}, functional_groups = fg)
            
        elif mediator_name == "TXB2":
            f1, f2, f3, f4 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH"), get_functional_group("oxy")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            f4.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f2, f3], "oxy": [f4]})]}
            return FattyAcid("FA", 20, {5: "Z", 13: "E"}, functional_groups = fg)
            
        elif mediator_name == "TXB3":
            f1, f2, f3, f4 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH"), get_functional_group("oxy")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            f4.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f2, f3], "oxy": [f4]})]}
            return FattyAcid("FA", 20, {5: "Z", 13: "E", 17: "Z"}, functional_groups = fg)
            
        elif mediator_name in {"PGF2alpha", "PGF2-alpha", "PGF2 alpha"}:
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f2, f3]})]}
            return FattyAcid("FA", 20, {5: "Z", 13: "E"}, functional_groups = fg)
            
        elif mediator_name == "PGD2":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("oxo")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f2], "oxo": [f3]})]}
            return FattyAcid("FA", 20, {5: "Z", 13: "E"}, functional_groups = fg)
            
        elif mediator_name == "PGE2":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("oxo"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f3], "oxo": [f2]})]}
            return FattyAcid("FA", 20, {5: "Z", 13: "E"}, functional_groups = fg)
            
        elif mediator_name == "PGB2":
            f1, f2 = get_functional_group("OH"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 1, functional_groups = {"OH": [f2]})]}
            return FattyAcid("FA", 20, {5: "Z", 13: "E"}, functional_groups = fg)
            
        elif mediator_name == "15d-PGJ2":
            f1, f2 = get_functional_group("OH"), get_functional_group("oxo")
            f1.position = 15
            f2.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 1, functional_groups = {"oxo": [f2]})]}
            return FattyAcid("FA", 20, {5: "Z", 12: "E", 14: "E"}, functional_groups = fg)
        
        elif mediator_name == "PGJ2":
            f1, f2 = get_functional_group("OH"), get_functional_group("oxo")
            f1.position = 15
            f2.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 1, functional_groups = {"OH": [f2]})]}
            return FattyAcid("FA", 20, {5: "Z", 13: "E"}, functional_groups = fg)
        
        elif mediator_name in {"PGEM", "PGE-M"}:
            f1, f2, f3 = get_functional_group("oxo"), get_functional_group("oxo"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"oxo": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f3], "oxo": [f2]})]}
            return FattyAcid("FA", 20, {5: "Z"}, functional_groups = fg)
            
        elif mediator_name in {"PGF1alpha", "PGF1-alpha", "PGF1 alpha"}:
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f2, f3]})]}
            return FattyAcid("FA", 20, {13: "E"}, functional_groups = fg)
            
        elif mediator_name == "PDX":
            f1, f2 = get_functional_group("OH"), get_functional_group("OH")
            f1.position = 0
            f2.position = 17
            fg = {"OH": [f1, f2]}
            return FattyAcid("FA", 22, {4: "Z", 7: "Z", 11: "E", 13: "Z", 15: "E", 19: "Z"}, functional_groups = fg)
            
        elif mediator_name in {"Oleic acid", "OA"}:
            return FattyAcid("FA", 18, {9: "Z"})
        
        elif mediator_name == "DGLA":
            return FattyAcid("FA", 20, {8: "Z", 11: "Z", 14: "Z"})
        
        elif mediator_name in {"iPF2alpha-VI", "iPF2-alpha-VI", "iPF2 alpha-VI"}:
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("oxo"), get_functional_group("OH")
            f1.position = 5
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"oxo": [f2], "OH": [f3]})]}
            return FattyAcid("FA", 20, {6: "E", 14: "Z"}, functional_groups = fg)
            
        elif mediator_name == "PGE1":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("oxo"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f3], "oxo": [f2]})]}
            return FattyAcid("FA", 20, {13: "E"}, functional_groups = fg)
            
        elif mediator_name == "PGE3":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("oxo"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f3], "oxo": [f2]})]}
            return FattyAcid("FA", 20, {5: "Z", 13: "E", 17: "Z"}, functional_groups = fg)
        

            
        
        return None



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

class BaseParserEventHandler:

    def __init__(self):
        self.registered_events = {}
        self.rule_names = set()
        self.parser = None
        self.content = None
        self.debug = None
    
    
    # checking if all registered events are reasonable and orrur as rules in the grammar
    def sanity_check(self):
        
        for event_name in self.registered_events:
            if not event_name.endswith("_pre_event") and not event_name.endswith("_post_event"):
                raise Exception("Parser event handler error: event '%s' does not contain the suffix '_pre_event' or '_post_event'" % event_name);
            
            rule_name = event_name.replace("_pre_event", "").replace("_post_event", "")
            if rule_name not in self.rule_names:
                raise Exception("Parser event handler error: rule '%s' in event '%s' is not present in the grammar%s" % (rule_name, event_name, " '" + self.parser.grammar_name + "'" if self.parser != None else ""))
    
    
    def handle_event(self, event_name, node):
        if self.debug != None and self.debug == "full":
            reg_event = '*' if event_name in self.registered_events else ''
            print("%s%s: '%s'" % (event_name, reg_event, node.get_text()))
            
            
        if event_name in self.registered_events:
            if self.debug != "" and self.debug != None and self.debug != "full": print("%s: \"%s\"" % (event_name, node.get_text()))
            self.registered_events[event_name](node)
            



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.parser.BaseParserEventHandler import BaseParserEventHandler
from pygoslin.domain.LipidAdduct import LipidAdduct
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.Adduct import Adduct
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.FattyAcid import FattyAcid

from pygoslin.domain.LipidCompleteStructure import LipidCompleteStructure
from pygoslin.domain.LipidFullStructure import LipidFullStructure
from pygoslin.domain.LipidStructureDefined import LipidStructureDefined
from pygoslin.domain.LipidSnPosition import LipidSnPosition
from pygoslin.domain.LipidMolecularSpecies import LipidMolecularSpecies
from pygoslin.domain.LipidSpecies import LipidSpecies
from pygoslin.domain.LipidSpeciesInfo import LipidSpeciesInfo
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.HeadGroup import HeadGroup
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.LipidClass import *
from pygoslin.domain.Cycle import *

last_numbers = {'un': 1, 'hen': 1, 'do': 2, 'di': 2, 'tri': 3, 'buta': 4, 'but': 4, 'tetra': 4, 'penta': 5, 'pent': 5, 'hexa': 6, 'hex': 6, 'hepta': 7, 'hept': 7, 'octa': 8, 'oct': 8, 'nona': 9, 'non': 9}

second_numbers = {'deca': 10, 'dec': 10, 'eicosa': 20, 'eicos': 20 , 'cosa': 20, 'cos': 20, 'triaconta': 30, 'triacont': 30, 'tetraconta': 40, 'tetracont': 40, 'pentaconta': 50, 'pentacont': 50, 'hexaconta': 60, 'hexacont': 60, 'heptaconta': 70, 'heptacont': 70, 'octaconta': 80, 'octacont': 80, 'nonaconta': 90, 'nonacont': 90}

special_numbers = {'meth': 1, 'etha': 2, 'eth': 2,  'propa': 3, 'isoprop': 3, 'prop': 3, 'propi': 3, 'propio': 3, 'buta': 4, 'but': 4, 'butr': 4, 'furan': 5, 'valer': 5, 'eicosa': 20, 'eicos': 20, 'icosa': 20, 'icos': 20, 'prosta': 20, 'prost': 20, 'prostan': 20}

func_groups = {'keto': 'oxo', 'ethyl': 'Et', 'hydroxy': "OH", 'phospho': 'Ph', 'oxo': 'oxo', 'bromo': 'Br', 'methyl': 'Me', 'hydroperoxy': 'OOH', 'homo': '', 'Epoxy': 'Ep', 'fluro': 'F', 'fluoro': 'F', 'chloro': 'Cl', 'methylene': 'My', 'sulfooxy': 'Su', 'amino': 'NH2', 'sulfanyl': 'SH', 'methoxy': 'OMe', 'iodo': 'I', 'cyano': 'CN', 'nitro': 'NO2', 'OH': 'OH', 'thio': 'SH', 'mercapto': 'SH', 'carboxy': "COOH", 'acetoxy': 'Ac', 'cysteinyl': 'Cys', 'phenyl': 'Phe', 's-glutathionyl': "SGlu", 's-cysteinyl': "SCys", "butylperoxy": "BOO", "dimethylarsinoyl": "MMAs", "methylsulfanyl": "SMe", "imino": "NH", 's-cysteinylglycinyl': "SCG"}

ate = {'formate': 1, 'acetate': 2, 'butyrate': 4, 'propionate': 3, 'valerate': 5, 'isobutyrate': 4}

class FattyAcidParserEventHandler(BaseParserEventHandler):
    
    def __init__(self):
        super().__init__()
        
        self.registered_events["lipid_pre_event"] = self.reset_lipid
        self.registered_events["lipid_post_event"] = self.build_lipid
        self.registered_events["fatty_acid_post_event"] = self.set_fatty_acid
        self.registered_events["fatty_acid_recursion_post_event"] = self.set_fatty_acid
        
        self.registered_events["acid_single_type_pre_event"] = self.set_fatty_acyl_type
        self.registered_events["ol_ending_pre_event"] = self.set_fatty_acyl_type
        self.registered_events["acid_heavy_single_type_post_event"] = self.setting_fatty_acyl_type
        self.registered_events["double_bond_position_pre_event"] = self.set_double_bond_information
        self.registered_events["double_bond_position_post_event"] = self.add_double_bond_information
        self.registered_events["db_number_post_event"] = self.set_double_bond_position
        self.registered_events["cistrans_post_event"] = self.set_cistrans
        self.registered_events["acid_type_double_post_event"] = self.check_db
        self.registered_events["db_length_pre_event"] = self.open_db_length
        self.registered_events["db_length_post_event"] = self.close_db_length
        
        ## lengths
        self.registered_events["functional_length_pre_event"] = self.reset_length
        self.registered_events["fatty_length_pre_event"] = self.reset_length
        self.registered_events["functional_length_post_event"] = self.set_functional_length
        self.registered_events["fatty_length_post_event"] = self.set_fatty_length
        
        ## numbers
        self.registered_events["notation_specials_pre_event"] = self.special_number
        self.registered_events["notation_last_digit_pre_event"] = self.last_number
        self.registered_events["notation_second_digit_pre_event"] = self.second_number
        
        ## functional groups
        self.registered_events["functional_group_pre_event"] = self.set_functional_group
        self.registered_events["functional_group_post_event"] = self.add_functional_group
        self.registered_events["functional_pos_pre_event"] = self.set_functional_pos
        self.registered_events["functional_position_pre_event"] = self.set_functional_position
        self.registered_events["functional_group_type_pre_event"] = self.set_functional_type
        
        ## cyclo / epoxy
        self.registered_events["cyclo_position_pre_event"] = self.set_functional_group
        self.registered_events["cyclo_position_post_event"] = self.rearrange_cycle
        self.registered_events["epoxy_pre_event"] = self.set_functional_group
        self.registered_events["epoxy_post_event"] = self.add_epoxy
        self.registered_events["cycle_pre_event"] = self.set_cycle
        self.registered_events["methylene_post_event"] = self.set_methylene
        
        ## dioic
        self.registered_events["dioic_pre_event"] = self.set_functional_group
        self.registered_events["dioic_post_event"] = self.set_dioic
        self.registered_events["dioic_acid_pre_event"] = self.set_fatty_acyl_type
        self.registered_events["dial_post_event"] = self.set_dial
        
        
        ## prosta
        self.registered_events["prosta_pre_event"] = self.set_prosta
        self.registered_events["prosta_post_event"] = self.add_cyclo
        self.registered_events["reduction_pre_event"] = self.set_functional_group
        self.registered_events["reduction_post_event"] = self.reduction
        self.registered_events["homo_post_event"] = self.homo
        
        ## furan
        self.registered_events["tetrahydrofuran_pre_event"] = self.set_tetrahydrofuran
        self.registered_events["furan_pre_event"] = self.set_furan
        
        
        ## recursion
        self.registered_events["recursion_description_pre_event"] = self.set_recursion
        self.registered_events["recursion_description_post_event"] = self.add_recursion
        self.registered_events["recursion_pos_pre_event"] = self.set_recursion_pos
        self.registered_events["yl_ending_pre_event"] = self.set_yl_ending
        self.registered_events["acetic_acid_post_event"] = self.set_acetic_acid
        self.registered_events["acetic_recursion_pre_event"] = self.set_recursion
        self.registered_events["acetic_recursion_post_event"] = self.add_recursion
        
        self.registered_events["hydroxyl_number_pre_event"] = self.add_hydroxyl
        self.registered_events["ol_pre_event"] = self.setup_hydroxyl
        self.registered_events["ol_post_event"] = self.add_hydroxyls
        self.registered_events["ol_pos_post_event"] = self.set_yl_ending
        
        
        
        ## wax esters
        self.registered_events["wax_ester_pre_event"] = self.set_recursion
        self.registered_events["wax_ester_post_event"] = self.add_wax_ester
        self.registered_events["ate_post_event"] = self.set_ate
        self.registered_events["isoprop_post_event"] = self.set_iso
        self.registered_events["isobut_post_event"] = self.set_iso
        
        ## CoA
        self.registered_events["coa_post_event"] = self.set_coa
        self.registered_events["methyl_pre_event"] = self.set_methyl
        
        ## CAR
        self.registered_events["car_pre_event"] = self.set_car
        self.registered_events["car_post_event"] = self.add_car
        
        ## amine
        self.registered_events["ethanolamine_post_event"] = self.add_ethanolamine
        self.registered_events["amine_n_pre_event"] = self.set_recursion
        self.registered_events["amine_n_post_event"] = self.add_amine
        self.registered_events["amine_post_event"] = self.add_amine_name
        
        ## functional group position summary
        self.registered_events["fg_pos_summary_pre_event"] = self.set_functional_group
        self.registered_events["fg_pos_summary_post_event"] = self.add_summary
        self.registered_events["func_stereo_pre_event"] = self.add_func_stereo

        self.registered_events["nic_pre_event"] = self.set_nic
        self.registered_events["deuterium_number_pre_event"] = self.add_heavy_adduct
        
        self.debug = ""
        self.adduct = None



    def set_nic(self, node):
        self.tmp["nic"] = node.get_text()



    def add_heavy_adduct(self, node):
        if not self.adduct: self.adduct = Adduct("", "")
        self.adduct.heavy_elements[Element.H2] = int(node.get_text())

        
        
    def reset_lipid(self, node):
        self.level = LipidLevel.FULL_STRUCTURE
        self.headgroup = ""
        self.fatty_acyl_stack = [FattyAcid("FA")]
        self.tmp = {"fa1": {}}
        self.adduct = None
        
        
        
    def set_car(self, node):
        self.tmp["fg_pos"] = []
        self.tmp["fg_type"] = ""
        
        
        
    def add_ol(self, node):
        self.headgroup = "FOH"
        
        
        
    def homo(self, node):
        self.tmp["post_adding"] = list(p[0] for p in self.tmp["fg_pos"])
        
        
        
    def add_summary(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        self.tmp[fa_i]["fg_pos_summary"] = {k: v.upper() for k, v in self.tmp["fg_pos"]}
        
        
        
    def set_acetic_acid(self, node):
        self.fatty_acyl_stack[-1].num_carbon += 2
        self.headgroup = "FA"
        
        
        
    def add_func_stereo(self, node):
        self.tmp["fg_pos"][-1][1] = node.get_text()
        
        
        
    def set_yl_ending(self, node):
        l = int(node.get_text()) - 1
        if l == 0: return
    
        curr_fa = self.fatty_acyl_stack[-1]
        
        if "furan" in self.tmp:
            curr_fa.num_carbon -= l
            return
        
        
        if l == 1:
            fname = "Me"
            fg = get_functional_group(fname)
        elif l == 2:
            fname = "Et"
            fg = get_functional_group(fname)
        else:
            fa = FattyAcid("FA", num_carbon = l)
            # shift functional groups
            for fg_name, fg_list in curr_fa.functional_groups.items():
                remove_item = []
                for i, func_group in enumerate(fg_list):
                    if func_group.position <= l:
                        remove_item.append(i)
                        if fg_name not in fa.functional_groups: fa.functional_groups[fg_name] = []
                        func_group.position = l + 1 - func_group.position
                        fa.functional_groups[fg_name].append(func_group)
                for i in remove_item[::-1]:
                    del curr_fa.functional_groups[fg_name][i]
            curr_fa.functional_groups = {k: v for k, v in curr_fa.functional_groups.items() if len(v) > 0}
            
            #shift double bonds
            if type(curr_fa.double_bonds) == dict:
                fa.double_bonds = {l + 1 - k: v for k, v in curr_fa.double_bonds.items() if k <= l}
                for k in list(k for k in curr_fa.double_bonds if k <= l): del curr_fa.double_bonds[k]
            
            fname = "cc"
            fg = CarbonChain(fa)
            
        curr_fa.num_carbon -= l
        fg.position = l
        curr_fa.shift_positions(-l)
        if fname not in curr_fa.functional_groups: curr_fa.functional_groups[fname] = []
        curr_fa.functional_groups[fname].append(fg)
        
        
        
    def set_methylene(self, node):
        self.tmp["fg_type"] = "methylene"
        if len(self.tmp["fg_pos"]) > 1:
            if self.tmp["fg_pos"][0][0] < self.tmp["fg_pos"][1][0]: self.tmp["fg_pos"][1][0] += 1
            elif self.tmp["fg_pos"][0][0] > self.tmp["fg_pos"][1][0]: self.tmp["fg_pos"][0][0] += 1
            self.fatty_acyl_stack[-1].num_carbon += 1
            self.tmp["add_methylene"] = True
     
    
    def check_db(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        curr_fa = self.fatty_acyl_stack[-1]
        if "fg_pos_summary" in self.tmp[fa_i]:
            if type(curr_fa.double_bonds) != dict: curr_fa.double_bonds = {}
            for k, v in self.tmp[fa_i]["fg_pos_summary"].items():
                if v in {"E", "Z", ""} and k > 0 and k not in curr_fa.double_bonds: curr_fa.double_bonds[k] = v
        
        
    def add_car(self, node):
        self.headgroup = "CAR"
        
        
        
    def add_hydroxyl(self, node):
        self.tmp["hydroxyl_pos"].append(int(node.get_text()))
    
    
    
    def set_coa(self, node):
        self.headgroup = "CoA"
        
        
        
    def set_methyl(self, node):
        self.fatty_acyl_stack[-1].num_carbon += 1
    
    
    
    def setup_hydroxyl(self, node):
        self.tmp["hydroxyl_pos"] = []
        
        
        
    def set_iso(self, node):
        curr_fa = self.fatty_acyl_stack[-1]
        curr_fa.num_carbon -= 1
        fg = get_functional_group("Me")
        fg.position = 2
        if "Me" not in curr_fa.functional_groups: curr_fa.functional_groups["Me"] = []
        curr_fa.functional_groups["Me"].append(fg)
        
        
        
    def set_ate(self, node):
        self.fatty_acyl_stack[-1].num_carbon += ate[node.get_text()]
        self.headgroup = "WE"
        
        
        
    def add_amine(self, node):
        fa = self.fatty_acyl_stack.pop()
        fa.lipid_FA_bond_type = LipidFaBondType.AMIDE
        self.fatty_acyl_stack[-1].lipid_FA_bond_type = LipidFaBondType.AMIDE
        self.fatty_acyl_stack.insert(0, fa)
        
        
        
    def add_amine_name(self, node):
        self.headgroup = "NA"
        
        
        
    def add_ethanolamine(self, node):
        self.headgroup = "NAE"
        
        
    
    def add_hydroxyls(self, node):
        if len(self.tmp["hydroxyl_pos"]) > 1:
            fg_oh = get_functional_group("OH")
            for pos in sorted(self.tmp["hydroxyl_pos"], reverse = True)[:-1]:
                fg_insert = fg_oh.copy()
                fg_insert.position = pos
                if "OH" not in self.fatty_acyl_stack[-1].functional_groups: self.fatty_acyl_stack[-1].functional_groups["OH"] = []
                self.fatty_acyl_stack[-1].functional_groups["OH"].append(fg_insert)
        
        
        
    def set_double_bond_position(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        pos = int(node.get_text())
        self.tmp[fa_i]["db_position"] = pos - (sum([1 for p in self.tmp["reduction"] if p < pos]) if "reduction" in self.tmp else 0)
        
        
        
    def set_recursion_pos(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        self.tmp[fa_i]["recursion_pos"] = int(node.get_text())
        
        
        
    def set_recursion(self, node):
        self.tmp["fg_pos"] = []
        self.tmp["fg_type"] = ""
        self.fatty_acyl_stack.append(FattyAcid("FA"))
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        self.tmp[fa_i] = {"recursion_pos": 0}
        
        
        
    def add_recursion(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        pos = self.tmp[fa_i]["recursion_pos"]
        
        
        fa = self.fatty_acyl_stack.pop()
        fa.position = pos
        curr_fa = self.fatty_acyl_stack[-1]
        
        if "cyclo_yl" in self.tmp:
            fname = "cyclo"
            del self.tmp["cyclo_yl"]
        else:
            fname = self.headgroup
        if fname not in curr_fa.functional_groups: curr_fa.functional_groups[fname] = []
        curr_fa.functional_groups[fname].append(fa)
        
        self.tmp["added_func_group"] = True
        
        
        
    def add_wax_ester(self, node):
        fa = self.fatty_acyl_stack.pop()
        fa.lipid_FA_bond_type = LipidFaBondType.ETHER
        self.fatty_acyl_stack.insert(0, fa)
        
        
        
    def set_cycle(self, node):
        self.tmp["cyclo"] = True
        
        
        
    def set_fatty_acid(self, node):
        def switch_position(func_group, switch):
            func_group.position = switch - func_group.position
            for fg_name, fg_list in func_group.functional_groups.items():
                for fg in fg_list:
                    switch_position(fg, switch)
                    
        if "length_pattern" in self.tmp:
            l, d, num, length_pattern = 0, 0, self.tmp["length_tokens"], self.tmp["length_pattern"]
            if length_pattern in {"L", "S"}:
                l += num[0]
                
            elif length_pattern == "LS":
                l += num[0] + num[1]
            
            
            elif length_pattern in {"LL", "SL", "SS"}:
                l += num[0]
                d += num[1]
                
            elif length_pattern in {"LSL", "LSS"}:
                l += num[0] + num[1]
                d += num[2]
                
            elif length_pattern == "LSLS":
                l += num[0] + num[1]
                d += num[2] + num[3]
                
            elif length_pattern == "SLS":
                l += num[0]
                d += num[1] + num[2]
                
            elif len(length_pattern) > 0 and length_pattern[0] == "X":
                l += num[0]
                d += sum(num[1:])
            
            elif length_pattern == "LLS": # false
                raise RuntimeException("Cannot determine fatty acid and double bond length in '%s'" % node.get_text())
            
            curr_fa = self.fatty_acyl_stack[-1]
            curr_fa.num_carbon += l
            if type(curr_fa.double_bonds) == int: curr_fa.double_bonds = d
        
        
        if "noyloxy" in curr_fa.functional_groups:
            if self.headgroup == "FA": self.headgroup = "FAHFA"
            
            while len(curr_fa.functional_groups["noyloxy"]) > 0:
                fa = curr_fa.functional_groups["noyloxy"].pop()
            
                
                acyl = AcylAlkylGroup(fa)
                acyl.position = fa.position
                
                if "acyl" not in curr_fa.functional_groups: curr_fa.functional_groups["acyl"] = []
                curr_fa.functional_groups["acyl"].append(acyl)
            del curr_fa.functional_groups["noyloxy"]
            
            
        elif "nyloxy" in curr_fa.functional_groups or "yloxy" in curr_fa.functional_groups:
            yloxy = "nyloxy" if "nyloxy" in curr_fa.functional_groups else "yloxy"
            
            while len(curr_fa.functional_groups[yloxy]) > 0:
                fa = curr_fa.functional_groups[yloxy].pop()
            
                
                alkyl = AcylAlkylGroup(fa, alkyl = True)
                alkyl.position = fa.position
                
                if "alkyl" not in curr_fa.functional_groups: curr_fa.functional_groups["alkyl"] = []
                curr_fa.functional_groups["alkyl"].append(alkyl)
            del curr_fa.functional_groups[yloxy]
            
                
        elif sum([k[-2:] == "yl" for k in curr_fa.functional_groups]) > 0:
            while True:
                try:
                    yl = [k for k in curr_fa.functional_groups if k[-2:] == "yl"][0]
                except Exception:
                    break
            
                while len(curr_fa.functional_groups[yl]) > 0:
                    fa = curr_fa.functional_groups[yl].pop()
                    if "cyclo" in self.tmp:
                        
                        cyclo_len = curr_fa.num_carbon
                        self.tmp["cyclo_len"] = cyclo_len
                        if fa.position != cyclo_len and "furan" not in self.tmp: switch_position(curr_fa, 2 + cyclo_len)
                        fa.shift_positions(cyclo_len)
                        if "furan" in self.tmp: curr_fa.shift_positions(-1)
                        
                        for fg, fg_list in fa.functional_groups.items():
                            if fg not in curr_fa.functional_groups: curr_fa.functional_groups[fg] = fg_list
                            else: curr_fa.functional_groups[fg] += fg_list
                            
                        
                        curr_fa.num_carbon = cyclo_len + fa.num_carbon
                        
                        if type(curr_fa.double_bonds) == int: curr_fa.double_bonds = {}
                        if type(fa.double_bonds) == dict:
                            for pos, ez in fa.double_bonds.items():
                                curr_fa.double_bonds[pos + cyclo_len] = ez
                        
                        if "furan" in self.tmp and "tetrahydrofuran" not in self.tmp:
                            if type(curr_fa.double_bonds) == int:
                                curr_fa.double_bonds += 2
                            else:
                                curr_fa.double_bonds[1] = "E"
                                curr_fa.double_bonds[3] = "E"
                        
                        self.tmp["cyclo_yl"] = True
                        
                    else:
                        ## add carbon chains here here
                        ## special chains: i.e. ethyl, methyl
                        fg_name = ""
                        if (fa.double_bonds if type(fa.double_bonds) == int else len(fa.double_bonds)) == 0 and len(fa.functional_groups) == 0:
                            if fa.num_carbon == 1:
                                fg_name = "Me"
                                fg = get_functional_group(fg_name)
                            
                            elif fa.num_carbon == 2:
                                fg_name = "Et"
                                fg = get_functional_group(fg_name)
                                
                            if len(fg_name) > 0:
                                fg.position = fa.position
                                if fg_name not in curr_fa.functional_groups: curr_fa.functional_groups[fg_name] = []
                                curr_fa.functional_groups[fg_name].append(fg)
                            
                        if len(fg_name) == 0:
                            cc = CarbonChain(fa, position = fa.position)
                            if "cc" not in curr_fa.functional_groups: curr_fa.functional_groups["cc"] = []
                            curr_fa.functional_groups["cc"].append(cc)
                            
                if "cyclo" in self.tmp: del self.tmp["cyclo"]
                del curr_fa.functional_groups[yl]
           
           
        if "cyclo" in curr_fa.functional_groups:
            fa = curr_fa.functional_groups["cyclo"][0]
            del curr_fa.functional_groups["cyclo"]
            if "cyclo_len" not in self.tmp: self.tmp["cyclo_len"] = 5
            start_pos, end_pos = curr_fa.num_carbon + 1, curr_fa.num_carbon + self.tmp["cyclo_len"]
            fa.shift_positions(start_pos - 1)
            
            if "cy" in curr_fa.functional_groups:
                for cy in curr_fa.functional_groups["cy"]:
                    cy.shift_positions(start_pos - 1)
            
            for fg, fg_list in fa.functional_groups.items():
                if fg not in curr_fa.functional_groups: curr_fa.functional_groups[fg] = fg_list
                else: curr_fa.functional_groups[fg] += fg_list
            
            
            if type(curr_fa.double_bonds) == int: curr_fa.double_bonds = {}
            if type(fa.double_bonds) == dict:
                for pos, ez in fa.double_bonds.items():
                    curr_fa.double_bonds[pos + start_pos - 1] = ez
            if "furan" in self.tmp and "tetrahydrofuran" not in self.tmp:
                if type(curr_fa.double_bonds) == int:
                    curr_fa.double_bonds += 2
                else:
                    curr_fa.double_bonds[1 + curr_fa.num_carbon] = "E"
                    curr_fa.double_bonds[3 + curr_fa.num_carbon] = "E"
                    
            curr_fa.num_carbon += fa.num_carbon
                    
            self.tmp["fg_pos"] = [[start_pos, ""], [end_pos, ""]]
            self.add_cyclo(node)
            
            if "cyclo_len" in self.tmp: del self.tmp["cyclo_len"]
            if "cyclo" in self.tmp: del self.tmp["cyclo"]
            
            
        elif "cyclo" in self.tmp:
            self.tmp["cyclo_yl"] = True
            self.tmp["cyclo_len"] = curr_fa.num_carbon
            self.tmp["fg_pos"] = [[1, ""], [curr_fa.num_carbon, ""]]
            del self.tmp["cyclo"]
        
        
        self.tmp["length_pattern"] = ""
        self.tmp["length_tokens"] = []
        self.tmp["add_lengths"] = False
            
        
    def set_double_bond_information(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        self.tmp[fa_i]["db_position"] = 0
        self.tmp[fa_i]["db_cistrans"] = ""
        
        
        
    def reduction(self, node):
        self.fatty_acyl_stack[-1].num_carbon -= len(self.tmp["fg_pos"])
        for fg, fg_list in self.fatty_acyl_stack[-1].functional_groups.items():
            for func_group in fg_list:
                func_group.shift_positions(-len(self.tmp["fg_pos"]))
        self.tmp["reduction"] = [p[0] for p in self.tmp["fg_pos"]]
        
        
        
    def add_double_bond_information(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        pos = self.tmp[fa_i]["db_position"]
        cistrans = self.tmp[fa_i]["db_cistrans"]
        if cistrans == "" and "fg_pos_summary" in self.tmp[fa_i] and pos in self.tmp[fa_i]["fg_pos_summary"]: cistrans = self.tmp[fa_i]["fg_pos_summary"][pos]
        if pos == 0: return
        
        cistrans = cistrans.upper();
        del self.tmp[fa_i]["db_position"]
        del self.tmp[fa_i]["db_cistrans"]
        if type(self.fatty_acyl_stack[-1].double_bonds) == int: self.fatty_acyl_stack[-1].double_bonds = {}
        
        
        if pos not in self.fatty_acyl_stack[-1].double_bonds or len(self.fatty_acyl_stack[-1].double_bonds[pos]) == 0:
            self.fatty_acyl_stack[-1].double_bonds[pos] = cistrans
        
        
        
    def set_dioic(self, node):
        self.headgroup = "FA"
        
        pos = self.tmp["fg_pos"][1][0] if len(self.tmp["fg_pos"]) == 2 else self.fatty_acyl_stack[-1].num_carbon
        if "reduction" in self.tmp: pos -= len(self.tmp["reduction"])
        self.fatty_acyl_stack[-1].num_carbon -= 1
        func_group = get_functional_group("COOH")
        func_group.position = pos - 1
        if "COOH" not in self.fatty_acyl_stack[-1].functional_groups: self.fatty_acyl_stack[-1].functional_groups["COOH"] = []
        self.fatty_acyl_stack[-1].functional_groups["COOH"].append(func_group)
        
        
        
    def set_cistrans(self, node):
        self.tmp["fa%i" % len(self.fatty_acyl_stack)]["db_cistrans"] = node.get_text()



    def setting_fatty_acyl_type(self, node):
        class texter:
            def __init__(self, t):
                self.t = t
            def get_text(self):
                return self.t
        self.set_fatty_acyl_type(texter(self.tmp["nic"] + " acid"))

        
        
    def set_fatty_acyl_type(self, node):
        t = node.get_text()
        
        if t[-2:] == "ol": self.headgroup = "FOH"
        elif t in {"noic acid", "nic acid", "dioic_acid"}: self.headgroup = "FA"
        elif t in {"nal", "dial"}: self.headgroup = "FAL"
        elif t in {"acetate", "noate", "nate"}: self.headgroup = "WE"
        elif t == "ne":
            self.headgroup = "HC"
            self.fatty_acyl_stack[-1].lipid_FA_bond_type = LipidFaBondType.ETHER
            
        else: self.headgroup = t
        
        
        
    def set_lipid_level(self, level):
        self.level = self.level if self.level.value < level.value else level
        
        
        
    def set_dial(self, node):
        curr_fa = self.fatty_acyl_stack[-1]
        pos = curr_fa.num_carbon
        fg = get_functional_group("oxo")
        fg.position = pos
        if "oxo" not in curr_fa.functional_groups: curr_fa.functional_groups["oxo"] = []
        curr_fa.functional_groups["oxo"].append(fg)
        
        
    
    def reset_length(self, node):
        self.tmp["length"] = 0
        self.tmp["length_pattern"] = ""
        self.tmp["length_tokens"] = []
        self.tmp["add_lengths"] = True
    
    
    
    def set_functional_length(self, node):
        if self.tmp["length"] != len(self.tmp["fg_pos"]):
            raise LipidException("Length of functional group '%i' does not match with number of its positions '%i'" % (self.tmp["length"], len(self.tmp["fg_pos"])))
    
    
    
    def set_fatty_length(self, node):
        self.tmp["add_lengths"] = False
        
        
        
    def rearrange_cycle(self, node):
        if "post_adding" in self.tmp:
            self.fatty_acyl_stack[-1].num_carbon += len(self.tmp["post_adding"])
            del self.tmp["post_adding"]
            
        curr_fa = self.fatty_acyl_stack[-1]
        start = self.tmp["fg_pos"][0][0]
        if "cy" in curr_fa.functional_groups:
            for cy in curr_fa.functional_groups["cy"]:
                shift = start - cy.position
                if shift == 0: continue
                cy.rearrange_functional_groups(curr_fa, shift)
        
        
        
    def add_cyclo(self, node):
        
        start = self.tmp["fg_pos"][0][0]
        end = self.tmp["fg_pos"][1][0]
            
        cyclo_db = None
        # check double bonds
        if type(self.fatty_acyl_stack[-1].double_bonds) == dict and len(self.fatty_acyl_stack[-1].double_bonds) > 0:
            cyclo_db = {db_pos: val for db_pos, val in self.fatty_acyl_stack[-1].double_bonds.items() if start <= db_pos <= end}
            
            for pos in cyclo_db:
                del self.fatty_acyl_stack[-1].double_bonds[pos]
                
        # check functional_groups
        cyclo_fg, remove_list, curr_fa = {}, set(), self.fatty_acyl_stack[-1]
        
        if "noyloxy" in curr_fa.functional_groups:
            remove_item = []
            for i, func_group in enumerate(curr_fa.functional_groups["noyloxy"]):
                if start <= func_group.position <= end:
                    cc = CarbonChain(func_group, position = func_group.position)
                    
                    if "cc" not in curr_fa.functional_groups: curr_fa.functional_groups["cc"] = []
                    curr_fa.functional_groups["cc"].append(cc)
                    remove_item.append(i)
                    
            for i in remove_item[::-1]: del curr_fa.functional_groups["noyloxy"][i]
            if len(curr_fa.functional_groups["noyloxy"]) == 0: remove_list.add("noyloxy")
            
        for fg, fg_list in curr_fa.functional_groups.items():
            remove_item = []
            
            for i, func_group in enumerate(fg_list):
                if start <= func_group.position <= end:
                    if fg not in cyclo_fg: cyclo_fg[fg] = []
                    cyclo_fg[fg].append(func_group)
                    remove_item.append(i)
                    
            for i in remove_item[::-1]: del curr_fa.functional_groups[fg][i]
            if len(fg_list) == 0: remove_list.add(fg)
            
        for fg in remove_list: del curr_fa.functional_groups[fg]
        
        bridge_chain = []
        if "furan" in self.tmp:
            del self.tmp["furan"]
            bridge_chain = [Element.O]
        
        cycle = Cycle(end - start + 1 + len(bridge_chain), start = start, end = end, double_bonds = cyclo_db, functional_groups = cyclo_fg, bridge_chain = bridge_chain)
        if "cy" not in self.fatty_acyl_stack[-1].functional_groups: self.fatty_acyl_stack[-1].functional_groups["cy"] = []
        self.fatty_acyl_stack[-1].functional_groups["cy"].append(cycle)
        
        
        
    def add_epoxy(self, node):
        self.tmp["fg_pos"] = self.tmp["fg_pos"][:1]
        self.tmp["fg_type"] = "Epoxy"
        
    
    
    def special_number(self, node):
        if self.tmp["add_lengths"]:
            self.tmp["length"] += special_numbers[node.get_text()]
            self.tmp["length_pattern"] += "X"
            self.tmp["length_tokens"].append(special_numbers[node.get_text()])
        
        
        
    def last_number(self, node):
        if self.tmp["add_lengths"]:
            self.tmp["length"] += last_numbers[node.get_text()]
            self.tmp["length_pattern"] += "L"
            self.tmp["length_tokens"].append(last_numbers[node.get_text()])
        
        
        
    def second_number(self, node):
        if self.tmp["add_lengths"]:
            self.tmp["length"] += second_numbers[node.get_text()]
            self.tmp["length_pattern"] += "S"
            self.tmp["length_tokens"].append(second_numbers[node.get_text()])
        
        
        
    def open_db_length(self, node):
        self.tmp["add_lengths"] = True
        
        
        
    def close_db_length(self, node):
        self.tmp["add_lengths"] = False
        
        
        
    def set_functional_group(self, node):
        self.tmp["fg_pos"] = []
        self.tmp["fg_type"] = ""
        
        
        
    def set_prosta(self, node):
        minus_pos = (sum([1 for p in self.tmp["reduction"] if p < 8]) if "reduction" in self.tmp else 0)
        self.tmp["fg_pos"] = [[8 - minus_pos, ""], [12 - minus_pos, ""]]
        self.tmp["fg_type"] = "cy"
        
        
    def set_tetrahydrofuran(self, node):
        self.tmp["furan"] = True
        self.tmp["tetrahydrofuran"] = True
        self.set_cycle(node)
        
        
        
    def set_furan(self, node):
        self.tmp["furan"] = True
        self.set_cycle(node)
        
        
        
    def add_functional_group(self, node):
        if "added_func_group" in self.tmp: 
            del self.tmp["added_func_group"]
            return
        
        elif "add_methylene" in self.tmp: 
            del self.tmp["add_methylene"]
            self.add_cyclo(node)
            return
        
        t = self.tmp["fg_type"]
        
        if t != "acetoxy":
            if t not in func_groups: raise LipidException("Unknown functional group: '%s'" % t)
            t = func_groups[t]
            if len(t) == 0: return
            fg = get_functional_group(t)
        else:
            fg = AcylAlkylGroup(FattyAcid("O", num_carbon = 2))
        
        if t not in self.fatty_acyl_stack[-1].functional_groups: self.fatty_acyl_stack[-1].functional_groups[t] = []
        for pos in self.tmp["fg_pos"]:
            fg_insert = fg.copy()
            fg_insert.position = pos[0] - (sum([1 for p in self.tmp["reduction"] if p < pos[0]]) if "reduction" in self.tmp else 0)
            self.fatty_acyl_stack[-1].functional_groups[t].append(fg_insert)
        
        
    def set_functional_position(self, node):
        self.tmp["fg_pos"].append([0, ""])
        
        
        
    def set_functional_pos(self, node):
        self.tmp["fg_pos"][-1][0] = int(node.get_text())
        
        
        
    def set_functional_type(self, node):
        self.tmp["fg_type"] = node.get_text()
        
        
        
    def build_lipid(self, node):
        
        if "cyclo_yl" in self.tmp:
            self.tmp["fg_pos"] = [[1, ""], [self.tmp["cyclo_len"], ""]]
            self.add_cyclo(node)
            del self.tmp["cyclo_yl"]
            del self.tmp["cyclo_len"]
            
        
        if "post_adding" in self.tmp:
            def add_position(func_group, pos):
                func_group.position += func_group.position >= pos
                if type(func_group) == Cycle:
                    func_group.start += func_group.start >= pos
                    func_group.end += func_group.end >= pos
                for fg_name, fg_list in func_group.functional_groups.items():
                    for fg in fg_list:
                        add_position(fg, pos)
            curr_fa = self.fatty_acyl_stack[-1]
            curr_fa.num_carbon += len(self.tmp["post_adding"])
            for pos in self.tmp["post_adding"]:
                add_position(curr_fa, pos)
                if type(curr_fa.double_bonds) == dict:
                    curr_fa.double_bonds = {(k + (k >= pos)): v for k, v in curr_fa.double_bonds.items()}
        
        if type(self.fatty_acyl_stack[-1].double_bonds) == dict and len(self.fatty_acyl_stack[-1].double_bonds) > 0:
            if sum(len(ct) > 0 for p, ct in self.fatty_acyl_stack[-1].double_bonds.items()) != len(self.fatty_acyl_stack[-1].double_bonds):
                self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        
        lipid_level_class = None
        if self.level == LipidLevel.COMPLETE_STRUCTURE: lipid_level_class = LipidCompleteStructure
        elif self.level == LipidLevel.FULL_STRUCTURE: lipid_level_class = LipidFullStructure
        elif self.level == LipidLevel.STRUCTURE_DEFINED: lipid_level_class = LipidStructureDefined
        elif self.level == LipidLevel.SN_POSITION: lipid_level_class = LipidSnPosition
        elif self.level == LipidLevel.MOLECULAR_SPECIES: lipid_level_class = LipidMolecularSpecies
        elif self.level == LipidLevel.SPECIES: lipid_level_class = LipidSpecies
        
        headgroup = HeadGroup(self.headgroup)
        
        self.content = LipidAdduct()
        self.content.adduct = self.adduct
        self.content.lipid = lipid_level_class(headgroup, self.fatty_acyl_stack)




"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


import unittest
import os

try:
    import pyximport
    pyximport.install(setup_args = {"script_args" : ["--force"]}, language_level = 3)
except:
    print("Warning: cython module is not installed, parsing performance will be lower since pure python code will be applied.")


from pygoslin.parser.Parser import *
from pygoslin.parser.GoslinParserEventHandler import GoslinParserEventHandler
from pygoslin.parser.LipidMapsParserEventHandler import LipidMapsParserEventHandler
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.LipidMolecularSpecies import LipidMolecularSpecies
from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from random import randint

lipid_parser = LipidParser()
swiss_lipids_parser = SwissLipidsParser()
goslin_parser = GoslinParser()
lipid_maps_parser = LipidMapsParser()
hmdb_parser = HmdbParser()

class ParserTest(unittest.TestCase):
    PARSER_QUOTE = '\''
    
    def test_lipid_parser(self):
        global lipid_parser
        
        lipid_name = "PE 16:1-12:0[M+H]1+"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE 16:1_12:0[M+H]1+"
        
        lipid_name = "PE O-16:1p/12:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE P-16:0/12:0"
        
        lipid_name = "PAT16 16:1/12:0/14:1/8:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PAT16 16:1/12:0/14:1/8:0"
        
        lipid_name = "SLBPA 16:1/12:0/14:1"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        
        assert lipid.get_lipid_string() == "SLBPA 16:1_12:0_14:1", "w: %s" % lipid.get_lipid_string()
        
        lipid_name = "MLCL 16:1/12:0/14:1"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "LCL 16:1_12:0_14:1"
        
        
        lipid_name = "DLCL 14:1/8:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "DLCL 14:1_8:0"
        
        
        lipid_name = "PE O-12:1p/10:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE P-12:0/10:0"
        
        
        lipid_name = "PE 12:1/10:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE 12:1/10:0"
        
        
        lipid_name = "PIP[3'] 17:0/20:4(5Z,8Z,11Z,14Z)"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PIP(3') 17:0/20:4(5Z,8Z,11Z,14Z)"
        
        lipid_name = "AC2SGL 12:0-14:1"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "AC2SGL 12:0_14:1"
        
        lipid_name = "NAPE 16:1(6Z)/12:0/14:1(5Z)"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE-N(FA 14:1(5Z)) 16:1(6Z)/12:0"
        
        lipid_name = "PE-NMe 12:1(6Z)/10:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE-NMe 12:1(6Z)/10:0"
        
        lipid_name = "PIMIP 12:0-14:1"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PIMIP 12:0_14:1"
        
        lipid_name = "LCDPDAG 24:1"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "LCDPDAG 24:1"
        
        lipid_name = "LPIMIP 10:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "LPIMIP 10:0"
        
        
        
    def test_extended_class(self):
        global lipid_parser
        
        lipid_name = "PE O-16:1a-12:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_extended_class() == "PE-O"
        
        lipid_name = "LPE O-16:2p"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_extended_class() == "LPE-P", "wrong: %s" % lipid.get_extended_class()
        
        lipid_name = "PC O-16:1p/12:0"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_extended_class() == "PC-P"
        
        lipid_name = "PC O-16:1p"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_extended_class() == "PC-P"
        
        lipid_name = "LPC O-16:1a"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_extended_class() == "LPC-O"
        
        lipid_name = "LPC O-16:1"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_extended_class() == "LPC-O"
        
        lipid_name = "LPC 16:1"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_extended_class() == "LPC"
        
        lipid_name = "PC 16:1/12:2"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_extended_class() == "PC"
        
        
        
    def test_oh_count(self):
        lipid_parser = LipidParser()
        lipid = lipid_parser.parse("Cer 36:1;2");
        ohCount = lipid.lipid.info.get_total_functional_group_count("OH")
        assert 2 == ohCount
        asdCount = lipid.lipid.info.get_total_functional_group_count("ASD")
        assert 0 == asdCount
        lipid = lipid_parser.parse("Cer d36:1");
        ohCount = lipid.lipid.info.get_total_functional_group_count("OH")
        assert 2 == ohCount
        lipid = lipid_parser.parse("Cer 18:1;2/18:0");
        ohCount = lipid.lipid.info.get_total_functional_group_count("OH")
        assert 2 == ohCount
        lipid = lipid_parser.parse("Cer d18:1/18:0");
        ohCount = lipid.lipid.info.get_total_functional_group_count("OH")
        assert 2 == ohCount
        lipid = lipid_parser.parse("Cer 18:1;(OH)2/18:0");
        ohCount = lipid.lipid.info.get_total_functional_group_count("OH")
        assert 2 == ohCount
        
        
    def test_lipid_maps(self):
        lipid_maps_parser = LipidMapsParser()
        
        for lipid_name_input, lipid_name_output in [["PA(16:1/12:0)", "PA 16:1/12:0"],
                           ["PC(O-14:0/0:0)", "LPC O-14:0/0:0"],
                           ["SQMG(16:1(11Z)/0:0)", "SQMG 16:1(11Z)/0:0"],
                           ["TG(13:0/22:3(10Z,13Z,16Z)/22:5(7Z,10Z,13Z,16Z,19Z))[iso6]", "TG 13:0/22:3(10Z,13Z,16Z)/22:5(7Z,10Z,13Z,16Z,19Z)"],
                           ["13R-HODE", "13R-HODE"],
                           ["CL(1'-[20:0/20:0],3'-[20:4(5Z,8Z,11Z,14Z)/18:2(9Z,12Z)])", "CL 20:0/20:0/20:4(5Z,8Z,11Z,14Z)/18:2(9Z,12Z)"],
                           ["PA(P-20:0/18:3(6Z,9Z,12Z))", "PA P-20:0/18:3(6Z,9Z,12Z)"],
                           ["M(IP)2C(t18:0/20:0(2OH))", "M(IP)2C(1) 18:0;3OH,4OH/20:0;2OH"],
                           ["Cer(d16:2(4E,6E)/22:0(2OH))", "Cer 16:2(4E,6E);1OH,3OH/22:0;2OH"],
                           ["MG(18:1(11E)/0:0/0:0)[rac]", "MG 18:1(11E)/0:0/0:0"],
                           ["PAT18(24:1(2E)(2Me,4Me[S],6Me[S])/25:1(2E)(2Me,4Me[S],6Me[S])/26:1(2E)(2Me,4Me[S],6Me[S])/24:1(2E)(2Me,4Me[S],6Me[S]))", "PAT18 24:1(2E);2Me,4Me,6Me/25:1(2E);2Me,4Me,6Me/26:1(2E);2Me,4Me,6Me/24:1(2E);2Me,4Me,6Me"],
                           ["(3'-sulfo)Galbeta-Cer(d18:1/20:0)", "SHexCer 18:1;O2/20:0"],
                           ["GlcCer(d15:2(4E,6E)/22:0(2OH))", "GlcCer(1) 15:2(4E,6E);3OH/22:0;2OH"]
                          ]:
            lipid = lipid_maps_parser.parse(lipid_name_input)
            assert lipid_maps_parser.word_in_grammar
            assert lipid != None
            assert lipid.get_lipid_string() == lipid_name_output, "wrong: %s" % lipid.get_lipid_string()
        
        
    @unittest.expectedFailure
    def test_LP(self):
        global lipid_parser
        lipid = lipid_parser.parse("LP 19:1p")
        
        
    def test_hydroxyls(self):
        global goslin_parser, swiss_lipids_parser, lipid_maps_parser, hmdb_parser
        
        lipid = swiss_lipids_parser.parse("Cer(d18:1(4E)/24:0-2OH)")
        assert lipid != None
        assert lipid.get_lipid_string() == "Cer 18:1(4);(OH)2/24:0;OH"
        assert lipid.get_sum_formula() == "C42H83NO4"
        assert abs(lipid.get_mass() - 665.632209) < 1e-3
        
        
        lipid = swiss_lipids_parser.parse("Cer(d18:1(4E)/24:0(2OH))")
        assert lipid != None
        assert lipid.get_lipid_string() == "Cer 18:1(4);(OH)2/24:0;OH"
        assert lipid.get_sum_formula() == "C42H83NO4"
        assert abs(lipid.get_mass() - 665.632209) < 1e-3
        
        
        lipid = lipid_maps_parser.parse("Cer(d18:1(4E)/24:0(2OH))")
        assert lipid != None
        assert lipid.get_lipid_string() == "Cer 18:1(4E);1OH,3OH/24:0;2OH"
        assert lipid.get_sum_formula() == "C42H83NO4"
        assert abs(lipid.get_mass() - 665.632209) < 1e-3
        
        
        lipid = goslin_parser.parse("Cer 18:1(4E);2/24:0;1")
        assert lipid != None
        assert lipid.get_lipid_string() == "Cer 18:1(4);(OH)2/24:0;OH"
        assert lipid.get_sum_formula() == "C42H83NO4"
        assert abs(lipid.get_mass() - 665.632209) < 1e-3
        
        lipid = hmdb_parser.parse("SM(d18:1(4)/16:1(9Z)(OH))")
        assert lipid != None
        assert lipid.get_lipid_string() == "SM 18:1(4);OH/16:1(9);OH"
        assert lipid.get_sum_formula() == "C39H77N2O7P"
        assert abs(lipid.get_mass() - 716.546841) < 1e-3


    def test_lyso(self):
        global lipid_parser
        
        lipid_name = "LPA O-16:1a"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "LPA O-16:1", "w: %s" % lipid.get_lipid_string()
        
        lipid_name = "LPC O-16:1a"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "LPC O-16:1"
        
        lipid_name = "LPE O-16:1p"
        lipid = lipid_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "LPE P-16:0", "w: %s" % lipid.get_lipid_string()
            
            
            
    @unittest.expectedFailure
    def test_lpe_fail(self):
        global lipid_parser
        lipid_name = "LPE O-16:1p/12:0"
        lipid = lipid_parser.parse(lipid_name)
        
        
    @unittest.expectedFailure
    def test_lipid_parser_fail(self):
        global lipid_parser
        lipid_name = "fail"
        lipid = lipid_parser.parse(lipid_name)
        
        
        
    def test_species_level(self):
        global goslin_parser
        
        lipid_name = "PG 22:1(5Z)/12:0"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string(LipidLevel.FULL_STRUCTURE) == "PG 22:1(5Z)/12:0"
        assert lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "PG 22:1(5)/12:0"
        assert lipid.get_lipid_string(LipidLevel.SN_POSITION) == "PG 22:1/12:0"
        assert lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "PG 22:1_12:0"
        assert lipid.get_lipid_string(LipidLevel.SPECIES) == "PG 34:1"
        assert lipid.get_lipid_string(LipidLevel.CLASS) == "PG"
        assert lipid.get_lipid_string(LipidLevel.CATEGORY) == "GP"
        
        lipid_name = "Cer 28:1;2"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "Cer 28:1;O2"
        
        lipid_name = "DAG 38:1"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "DG 38:1"
        
        
        
        
    def test_info_level(self):
        global swiss_lipids_parser
        global lipid_maps_parser
        
        lipid_name = "PG(22:1(5Z)/12:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string(LipidLevel.FULL_STRUCTURE) == "PG 22:1(5Z)/12:0"
        assert lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "PG 22:1(5)/12:0"
        assert lipid.get_lipid_string(LipidLevel.SN_POSITION) == "PG 22:1/12:0"
        assert lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "PG 22:1_12:0"
        assert lipid.get_lipid_string(LipidLevel.SPECIES) == "PG 34:1"
        assert lipid.get_lipid_string(LipidLevel.CLASS) == "PG"
        assert lipid.get_lipid_string(LipidLevel.CATEGORY) == "GP"
        
        lipid_name = "PG(22:1(5Z)/12:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.lipid.info.level == LipidLevel.FULL_STRUCTURE
        assert lipid.get_lipid_string() == "PG 22:1(5Z)/12:0"
        
        lipid_name = "PG(22:1(5)/12:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid
        assert lipid.lipid.info.level == LipidLevel.STRUCTURE_DEFINED
        assert lipid.get_lipid_string() == "PG 22:1(5)/12:0"
        
        lipid_name = "PG(22:1/12:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid
        assert lipid.lipid.info.level == LipidLevel.SN_POSITION
        assert lipid.get_lipid_string() == "PG 22:1/12:0"
        
        lipid_name = "PG(22:1_12:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid
        assert lipid.lipid.info.level == LipidLevel.MOLECULAR_SPECIES
        assert lipid.get_lipid_string() == "PG 22:1_12:0"
        
        lipid_name = "LPG(O-22:1)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid
        assert lipid.lipid.info.level == LipidLevel.SPECIES
        assert lipid.get_lipid_string() == "LPG O-22:1"
        
        
        lipid_name = "PG(22:1(5Z)/12:0)"
        lipid = lipid_maps_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string(LipidLevel.FULL_STRUCTURE) == "PG 22:1(5Z)/12:0"
        assert lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "PG 22:1(5)/12:0"
        assert lipid.get_lipid_string(LipidLevel.SN_POSITION) == "PG 22:1/12:0"
        assert lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "PG 22:1_12:0"
        assert lipid.get_lipid_string(LipidLevel.SPECIES) == "PG 34:1"
        assert lipid.get_lipid_string(LipidLevel.CLASS) == "PG"
        assert lipid.get_lipid_string(LipidLevel.CATEGORY) == "GP"
        
        
        
        
        
    def test_lipid_names(self):
        file_name = os.path.join("pygoslin", "data", "goslin", "Goslin.g4")
        goslin_parser = Parser(GoslinParserEventHandler(), file_name, ParserTest.PARSER_QUOTE)
        
        ## glycerophospholipid
        lipid_name = "PE 16:1(3E)/12:0"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "PE 16:1(3)/12:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.SN_POSITION) == "PE 16:1/12:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "PE 16:1_12:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.SPECIES) == "PE 28:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.CLASS) == "PE"
        assert lipid.lipid.get_lipid_string(LipidLevel.CATEGORY) == "GP"
        
        ## sphingolipid
        lipid_name = "Cer 16:1(5Z);2/12:0"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "Cer 16:1(5);(OH)2/12:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.SN_POSITION) == "Cer 16:1;O2/12:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "Cer 16:1;O2/12:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.SPECIES) == "Cer 28:1;O2"
        assert lipid.lipid.get_lipid_string(LipidLevel.CLASS) == "Cer"
        assert lipid.lipid.get_lipid_string(LipidLevel.CATEGORY) == "SP"
        
        ## glycerolipid
        lipid_name = "TAG 16:1(5Z)/12:0/20:2(8Z,11Z)"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        
        assert lipid.lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "TG 16:1(5)/12:0/20:2(8,11)"
        assert lipid.get_lipid_string(LipidLevel.SN_POSITION) == "TG 16:1/12:0/20:2"
        assert lipid.lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "TG 16:1_12:0_20:2"
        assert lipid.lipid.get_lipid_string(LipidLevel.SPECIES) == "TG 48:3"
        assert lipid.lipid.get_lipid_string(LipidLevel.CLASS) == "TG"
        assert lipid.lipid.get_lipid_string(LipidLevel.CATEGORY) == "GL"
        
        
        
        
        ## sterol
        lipid_name = "ChE 16:1(4Z)"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        
        assert lipid.lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "SE 27:1/16:1(4)"
        assert lipid.lipid.get_lipid_string(LipidLevel.SN_POSITION) == "SE 27:1/16:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "SE 27:1/16:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.SPECIES) == "SE 27:1/16:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.CLASS) == "SE 27:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.CATEGORY) == "ST"
        
        ## sterol
        lipid_name = "SE 27:1/16:1(4Z)"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        
        assert lipid.lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "SE 27:1/16:1(4)"
        assert lipid.lipid.get_lipid_string(LipidLevel.SN_POSITION) == "SE 27:1/16:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "SE 27:1/16:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.SPECIES) == "SE 27:1/16:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.CLASS) == "SE 27:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.CATEGORY) == "ST"
        
        
        
        
        ## PC
        lipid_name = "PC O-18:1(4E)a/16:0"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        
        assert lipid.lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED) == "PC O-18:1(4)/16:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.SN_POSITION) == "PC O-18:1/16:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.MOLECULAR_SPECIES) == "PC O-18:1_16:0"
        assert lipid.lipid.get_lipid_string(LipidLevel.SPECIES) == "PC O-34:1"
        assert lipid.lipid.get_lipid_string(LipidLevel.CLASS) == "PC"
        assert lipid.lipid.get_lipid_string(LipidLevel.CATEGORY) == "GP"
        
        
    def test_adduct(self):
        goslin_parser_event_handler = GoslinParserEventHandler()
        file_name = os.path.join("pygoslin", "data", "goslin", "Goslin.g4")
        goslin_parser = Parser(goslin_parser_event_handler, file_name)

        lipid_name = "PE 16:1/12:0[M+H]1+"
        lipid = goslin_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE 16:1/12:0[M+H]1+"
        
        
        
        
      
    def test_swiss_lipids(self):
        global swiss_lipids_parser
        
        lipid_name = "TG(O-16:0/18:3(6Z,9Z,12Z)/18:1(11Z))"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "TG O-16:0/18:3(6Z,9Z,12Z)/18:1(11Z)"
        
        lipid_name = "PIP2[4,5](21:0/24:4(9Z,12Z,15Z,18Z))"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PIP2(4',5') 21:0/24:4(9Z,12Z,15Z,18Z)"
        
        lipid_name = "GalGb3Cer(d18:0/14:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "GalGb3Cer 18:0;OH/14:0", "wrong %s" % lipid.get_lipid_string()
        
        lipid_name = "PI(34:5(19Z,22Z,25Z,28Z,31Z)/18:1(6Z))"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PI 34:5(19Z,22Z,25Z,28Z,31Z)/18:1(6Z)"
        
        lipid_name = "PIP(22:6/20:4)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PIP 22:6/20:4"
        
        lipid_name = "NAPE (12:0/30:4(15Z,18Z,21Z,24Z)/12:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE-N(FA 12:0) 12:0/30:4(15Z,18Z,21Z,24Z)"
        
        lipid_name = "NAPE (12:0/0:0/12:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "LPE-N(FA 12:0) 12:0/0:0"
        
        lipid_name = "NAPE (30:5(12Z,15Z,18Z,21Z,24Z)/18:0/8:0)"
        lipid = swiss_lipids_parser.parse(lipid_name)
        assert lipid != None
        assert lipid.get_lipid_string() == "PE-N(FA 8:0) 30:5(12Z,15Z,18Z,21Z,24Z)/18:0"
        assert lipid.get_sum_formula() == "C61H110NO9P"
      
      
      
      
    
    def test_parser_read(self):
        global lipid_parser
        lipidnames = []
        file_name = os.path.join("pygoslin", "data", "goslin", "testfiles", "lipidnames.csv")
        with open(file_name, mode = "rt", encoding= "utf-8") as infile:
            for line in infile:
                line = line.strip().strip(" ")
                if len(line) < 2: continue
                if line[0] == "#": continue
                lipidnames.append(Parser.split_string(line, ",", "\"")[0].strip("\""))
        
        for lipid_name in lipidnames:
            try:
                lipid = lipid_parser.parse(lipid_name)
            except Exception as e:
                print("Lipid could not be parsed: '%s'" % lipid_name)
                raise e
            assert lipid != None
